unit BinaryBuffer;

interface

uses
	SysUtils;

type
	Natural = 0..High(Integer);

type
	TGapBuffer = class(TObject)
	public type
		TRead = function(var Buffer; Count: Cardinal): Cardinal of object;
		TWrite = function(const Buffer; Count: Cardinal): Cardinal of object;
		TWriter = class(TObject)
		private
			FBuffer: TGapBuffer;
			FData: Pointer;
			FSize: Natural;
			FCapacity: Natural;
		public
			constructor Create(ABuffer: TGapBuffer);
			destructor Destroy; override;
			procedure Write(const Data; Size: Integer);
		end;
	private
		FSize: Natural;
		FCapacity: Natural;
		FGapFirst: Natural;
		FGapLast: Natural;
		FData: Pointer;
		function GetData(Position: Natural): Byte;
		procedure MoveGap(Position: Natural);
	protected
		procedure Edit(Position: Natural; DeleteLength, InsertLength: Integer; const Data);
		procedure MoveBlock(Source, Dest: Natural; Size: Natural); virtual;
		procedure ReallocateBlock(NewCapacity: Natural;
			From1, To1, Size1, From2, To2, Size2, From3, To3, Size3: Natural); virtual;
		procedure InternalClear; virtual;
		procedure Read(Position: Natural; Size: Natural; out Data);
		procedure BeforeChange(First, Last: Natural); virtual;
		procedure AfterChange(First, OldLast, NewLast: Natural); virtual;
		property GapFirst: Natural read FGapFirst;
		property GapLast: Natural read FGapLast;
	public
		constructor Create;
		destructor Destroy; override;
		procedure Clear;
		procedure Insert(Position: Natural; InsertLength: Integer; const Data);
		procedure Delete(Position: Natural; DeleteLength: Integer);
		procedure Replace(Position: Natural; DeleteLength, InsertLength: Integer; const Data); virtual;
		function DirectAccess: PByte;
		procedure Save(Write: TWrite);
		procedure Load(Read: TRead; ExpectingSize: Integer = 0);
		property Size: Natural read FSize;
		property Capacity: Natural read FCapacity;
		property Data[Position: Natural]: Byte read GetData; default;
	end;

type
	TBufferAccessor = record
		Buffer: TGapBuffer;
		Index: Natural;
		function Take: Byte;
		function EndOfBuffer: Boolean; inline;
		function Match(C: Byte): Boolean;
		function Match2(C1, C2: Byte): Boolean;
		procedure Advance;
		procedure Advance2;
	end;

type
	TBuffer = class;

	PMark = ^TMark;
	TMark = record
		Next: PMark;
		First: Natural;
		Last: Natural;
	end;

	TBufferMarker = class(TObject)
	private
		FMark: PMark;
		FBuffer: TBuffer;
	protected
		procedure BufferChanged(First, OldLast, NewLast: Natural);
	public
		constructor Create(ABuffer: TBuffer);
		destructor Destroy; override;
		procedure Clear;
		procedure Mark(First, Last: Natural);
		function MarkedRange(First, Last: Natural; var Pinpoint: Boolean): Boolean;
		function Marked: Boolean;
		function FindPrior(Position: Natural): PMark;
		function FindNext(Position: Natural): PMark;
	end;

	PUndoBuffer = ^TUndoBuffer;
	TUndoBuffer = record
		Next: PUndoBuffer;
		Position: Natural;
		DeleteLength: Integer; {}
		InsertLength: Integer;
	end;

	TSubBufferRange = 0..1;

	TBuffer = class(TGapBuffer)
	public type
		TWriter = TGapBuffer.TWriter;
	private
		FSub: array[TSubBufferRange] of Pointer;
		FUndo: PUndoBuffer;
		FRedo: PUndoBuffer;
		FMarkers: array of TBufferMarker;
		FModified: TBufferMarker;
		FReadOnly: Boolean;
		procedure ClearUndoBuffer;
		procedure ClearRedoBuffer;
		procedure MakeUndoBuffer(var Result: PUndoBuffer; Action: PUndoBuffer);
		procedure AddMarker(AMarker: TBufferMarker);
		procedure RemoveMarker(AMarker: TBufferMarker);
		function GetSub(ASub: TSubBufferRange; Position: Natural): Byte;
		procedure SetSub(ASub: TSubBufferRange; Position: Natural; const Value: Byte);
	protected
		procedure InternalClear; override;
		procedure MoveBlock(Source, Dest: Natural; Size: Natural); override;
		procedure ReallocateBlock(NewCapacity: Natural;
			From1, To1, Size1, From2, To2, Size2, From3, To3, Size3: Natural); override;
		procedure AfterChange(First, OldLast, NewLast: Natural); override;
	public
		constructor Create;
		destructor Destroy; override;
		procedure Replace(Position: Natural; DeleteLength, InsertLength: Integer; const Data); override;
		function Undo: Natural;
		function Redo: Natural;
		function CanUndo: Boolean;
		function CanRedo: Boolean;
		function UndoPosition: Natural;
		function RedoPosition: Natural;
		function ModifiedRange(First, Last: Natural): Boolean;
		function Modified: Boolean;
		procedure ClearModified;
		function AllocateSubBuffer: TSubBufferRange;
		procedure ReleaseSubBuffer(AIndex: TSubBufferRange);
		procedure CheckReadOnly;
		property Sub[ASub: TSubBufferRange; Position: Natural]: Byte read GetSub write SetSub;
		property ReadOnly: Boolean read FReadOnly write FReadOnly;
	end;

type
	EBufferReadOnly = class(EAbort)
	end;

implementation

const
	InitialCapacity = 1024 * 4;

{ TGapBuffer }

procedure TGapBuffer.AfterChange(First, OldLast, NewLast: Natural);
begin
end;

procedure TGapBuffer.BeforeChange(First, Last: Natural);
begin
end;

procedure TGapBuffer.Clear;
var
	OldLast: Natural;
begin
  OldLast := FSize;
	BeforeChange(0, OldLast);
	InternalClear;
  AfterChange(0, OldLast, 0);
end;

constructor TGapBuffer.Create;
begin
	inherited;
end;

procedure TGapBuffer.Delete(Position: Natural; DeleteLength: Integer);
begin
	Replace(Position, DeleteLength, 0, nil^)
end;

destructor TGapBuffer.Destroy;
begin
	InternalClear;
	inherited;
end;

function TGapBuffer.DirectAccess: PByte;
begin
	Assert(FSize = FCapacity - (FGapLast - FGapFirst));

	if FGapFirst = FGapLast then
		Result := FData
	else
	begin
		if FGapFirst < FCapacity - FGapLast then
		begin
			MoveGap(0);
			Result := PByte(Cardinal(FData) + FGapLast);
		end
		else
		begin
			MoveGap(FSize);
			Result := FData;
		end
	end;
end;

procedure TGapBuffer.Edit(Position: Natural; DeleteLength, InsertLength: Integer; const Data);
var
	Start, Insert, Delete: Natural;
	NewCapacity, NewGapFirst, NewGapLast: Natural;
	EditFirst, EditLast: Natural;
	From1, From2, From3, To1, To2, To3, Size1, Size2, Size3: Natural;
	Diff, Growing: Integer;
	NewData: Pointer;
begin
	if DeleteLength >= 0 then
	begin
		Start := Position;
		Delete := DeleteLength;
	end
	else
	begin
		Start := Position + DeleteLength;
		Delete := - DeleteLength;
	end;

	Assert(Start + Delete <= FSize);

	BeforeChange(Start, Start + Delete);

	Insert := Abs(InsertLength);

	Diff := Insert - Delete;
	Inc(FSize, Diff);

	if Diff < 0 then
	begin
		EditFirst := Start + Insert;
		EditLast := Start + Delete;
		NewGapFirst := EditFirst;
		NewGapLast := EditLast + (FGapLast - FGapFirst);
		if EditLast <= FGapFirst then
		begin
			MoveBlock(EditLast, NewGapLast, FGapFirst - EditLast);
		end
		else if EditFirst >= FGapFirst then
		begin
			MoveBlock(FGapLast, FGapFirst, EditFirst - FGapFirst);
		end;
		FGapFirst := NewGapFirst;
		FGapLast := NewGapLast;
	end
	else if Diff > 0 then
	begin
		if Diff <= FGapLast - FGapFirst then
		begin
			MoveGap(Start + Delete);
			Inc(FGapFirst, Diff);
			Assert(FGapFirst <= FGapLast);
		end
		else
		begin
			Growing := FSize div 8;
			if Growing < 7 then
				Growing := 7
			else if Growing > 1024 * 1024 then
				Growing := 1024 * 1024;

			NewCapacity := (FSize + Growing) and not 3;
			
			NewGapFirst := Start;
			NewGapLast := NewCapacity - (FCapacity - (FGapLast - FGapFirst) - Start) + Delete;
			if NewGapFirst <= FGapFirst then
			begin
				From1 := 0;
				To1 := 0;
				Size1 := NewGapFirst;

				From2 := NewGapFirst + Delete;
				To2 := NewGapLast;
				Size2 := FGapFirst - (NewGapFirst + Delete);

				From3 := FGapLast;
				To3 := NewGapLast + Size2;
				Size3 := FCapacity - FGapLast;
			end
			else
			begin
				From1 := 0;
				To1 := 0;
				Size1 := FGapFirst;

				From2 := FGapLast;
				To2 := FGapFirst;
				Size2 := NewGapFirst - FGapFirst;

				From3 := From2 + Size2 + Delete;
				To3 := NewGapLast;
				Size3 := NewCapacity - NewGapLast;
			end;

			Assert(From1 + Size1 <= FCapacity);
			Assert(From2 + Size2 <= FCapacity);
			Assert(From3 + Size3 <= FCapacity);
			Assert(To1 + Size1 <= NewCapacity);
			Assert(To2 + Size2 <= NewCapacity);
			Assert(To3 + Size3 <= NewCapacity);

      ReallocateBlock(NewCapacity, From1, To1, Size1, From2, To2, Size2, From3, To3, Size3);

			FGapFirst := NewGapFirst + Insert;
			FGapLast := NewGapLast;
			FCapacity := NewCapacity;
		end
	end
	else
	begin
		MoveGap(Start + Insert);
	end;

	Assert(FSize = FCapacity - (FGapLast - FGapFirst));
	Assert(Start + Insert = FGapFirst);

	{݃GA͘ÄɂȂĂ͂}
	System.Move(Data, Pointer(Cardinal(FData) + Start)^, Insert);

	AfterChange(Start, Start + Delete, Start + Insert);
end;

function TGapBuffer.GetData(Position: Natural): Byte;
begin
	Assert(Position < Size);
	if Position < FGapFirst then
		Result := PByte(Cardinal(FData) + Position)^
	else
		Result := PByte(Cardinal(FData) + Position + Cardinal(FGapLast - FGapFirst))^
end;

procedure TGapBuffer.Insert(Position: Natural; InsertLength: Integer; const Data);
begin
	Replace(Position, 0, InsertLength, Data)
end;

procedure TGapBuffer.InternalClear;
begin
	FSize := 0;
	FCapacity := 0;
	FGapFirst := 0;
	FGapLast := 0;
	FreeMem(FData);
end;

procedure TGapBuffer.Load(Read: TRead; ExpectingSize: Integer);
var
	R, OldSize: Natural;
begin
	OldSize := FSize;

	BeforeChange(0, OldSize);
	InternalClear;

	FCapacity := InitialCapacity;
	if ExpectingSize > 0 then
		FCapacity := ExpectingSize + 1024;

	GetMem(FData, FCapacity);

	FSize := 0;

	repeat
		R := Read(Pointer(Cardinal(FData) + FSize)^, FCapacity - FSize);
		Inc(FSize, R);
		if FSize < FCapacity then Break;
		FCapacity := FCapacity * 2;
		ReallocMem(FData, FCapacity);
	until False;

	FGapFirst := FSize;
	FGapLast := FCapacity;

  ReallocateBlock(FCapacity, 0, 0, 0, 0, 0, 0, 0, 0, 0);

	AfterChange(0, OldSize, FSize);
end;

procedure TGapBuffer.MoveBlock(Source, Dest, Size: Natural);
begin
	Assert(Source + Size <= FCapacity);
	Assert(Dest + Size <= FCapacity);
	System.Move(
		Pointer(Cardinal(FData) + Source)^,
		Pointer(Cardinal(FData) + Dest)^,
		Size);
end;

procedure TGapBuffer.MoveGap(Position: Natural);
var
	GapSize: Natural;
begin
	GapSize := FGapLast - FGapFirst;

	if FGapFirst > Position then
		MoveBlock(Position, Position + GapSize, FGapFirst - Position)
	else if FGapFirst < Position then
		MoveBlock(FGapLast, FGapFirst, Position - FGapFirst);

	FGapFirst := Position;
	FGapLast := Position + GapSize;
end;

procedure TGapBuffer.Read(Position, Size: Natural; out Data);
var
	Size1: Integer;
begin
	if Position + Size <= FGapFirst then
	begin
		System.Move(Pointer(Cardinal(FData) + Position)^, Data, Size);
	end
	else if Position >= FGapFirst then
	begin
		System.Move(
			Pointer(Cardinal(FData) + Position + Cardinal(FGapLast - FGapFirst))^,
			Data, Size);
	end
	else
	begin
		Size1 := FGapFirst - Position;
		Assert(Size1 >= 0);
		System.Move(Pointer(Cardinal(FData) + Position)^, Data, Size1);
		System.Move(
			Pointer(Cardinal(FData) + FGapLast)^,
			Pointer(Cardinal(@Data) + Cardinal(Size1))^, Size - Size1);
	end
end;

procedure TGapBuffer.ReallocateBlock(NewCapacity: Natural;
	From1, To1, Size1, From2, To2, Size2, From3, To3, Size3: Natural);
var
	NewData: Pointer;
begin
	if FCapacity <> NewCapacity then
	begin
		GetMem(NewData, NewCapacity);
		System.Move(Pointer(Cardinal(FData) + From1)^, Pointer(Cardinal(NewData) + To1)^, Size1);
		System.Move(Pointer(Cardinal(FData) + From2)^, Pointer(Cardinal(NewData) + To2)^, Size2);
		System.Move(Pointer(Cardinal(FData) + From3)^, Pointer(Cardinal(NewData) + To3)^, Size3);
		FreeMem(FData);
		FData := NewData;
	end
end;

procedure TGapBuffer.Replace(Position: Natural; DeleteLength, InsertLength: Integer; const Data);
begin
	Edit(Position, DeleteLength, InsertLength, Data);
end;

procedure TGapBuffer.Save(Write: TWrite);
begin
	Write(FData^, FGapFirst);
	Write(Pointer(Cardinal(FData) + FGapLast)^, FCapacity - FGapLast);
end;

{ TBuffer.TWriter }

constructor TGapBuffer.TWriter.Create(ABuffer: TGapBuffer);
begin
	inherited Create;
	FBuffer := ABuffer;
	FCapacity := 1024;
	GetMem(FData, FCapacity);
end;

destructor TGapBuffer.TWriter.Destroy;
var
	OldSize: Natural;
begin
	OldSize := FBuffer.FSize;
	FBuffer.BeforeChange(0, OldSize);
	{Clear}
	FBuffer.InternalClear;
	{Set}
	FBuffer.FData := FData;
	FBuffer.FSize := FSize;
	FBuffer.FCapacity := FCapacity;
	FBuffer.FGapFirst := FSize;
	FBuffer.FGapLast := FCapacity;

	FBuffer.ReallocateBlock(FCapacity, 0, 0, 0, 0, 0, 0, 0, 0, 0);
	FBuffer.AfterChange(0, OldSize, FSize);

	inherited;
end;

procedure TGapBuffer.TWriter.Write(const Data; Size: Integer);
begin
	if Size > FCapacity - FSize then
	begin
		FCapacity := FCapacity * 2;
		ReallocMem(FData, FCapacity);
	end;
	System.Move(Data, Pointer(Cardinal(FData) + FSize)^, Size);
	Inc(FSize, Size);
end;

{ TBufferAccessor }

procedure TBufferAccessor.Advance;
begin
	Inc(Self.Index)
end;

procedure TBufferAccessor.Advance2;
begin
	Inc(Self.Index, 2)
end;

function TBufferAccessor.EndOfBuffer: Boolean;
begin
	Result := Self.Index >= Self.Buffer.Size
end;

function TBufferAccessor.Match(C: Byte): Boolean;
begin
	Result := (Self.Index < Self.Buffer.Size) and
		(Self.Buffer.Data[Self.Index] = C)
end;

function TBufferAccessor.Match2(C1, C2: Byte): Boolean;
begin
	Result := (Self.Index + 1 < Self.Buffer.Size) and
		(Self.Buffer.Data[Self.Index    ] = C1) and
		(Self.Buffer.Data[Self.Index + 1] = C2)
end;

function TBufferAccessor.Take: Byte;
begin
	Result := Self.Buffer.Data[Self.Index];
	Inc(Self.Index)
end;

{ TBufferMarker }

procedure TBufferMarker.BufferChanged(First, OldLast, NewLast: Natural);
var
	P, Pre: PMark;
begin
	if OldLast <> NewLast then
	begin
		Pre := nil;
		P := FMark;
		while P <> nil do
		begin
			if P^.First > OldLast then
			begin
				Inc(P^.First, NewLast - OldLast);
				Inc(P^.Last, NewLast - OldLast);
			end
			else if P^.First >= First then
			begin
				P^.First := First;
				if P^.Last <= OldLast then
					P^.Last := NewLast
				else
					Inc(P^.Last, NewLast - OldLast);
			end
			else if P^.Last >= First then
			begin
				Inc(P^.Last, NewLast - OldLast);
			end;

			if Pre <> nil then
			begin
				if Pre^.Last >= P^.First then
				begin
					Pre^.Last := P^.Last;
					Pre^.Next := P^.Next;
					FreeMem(P);
					P := Pre;
				end;
			end;

			Pre := P;
			P := P^.Next;
		end;
	end
end;

procedure TBufferMarker.Clear;
var
	R: PMark;
begin
	while FMark <> nil do
	begin
		R := FMark;
		FMark := FMark.Next;
		FreeMem(R);
	end;
end;

constructor TBufferMarker.Create(ABuffer: TBuffer);
begin
	inherited Create;
	FBuffer := ABuffer;
	ABuffer.AddMarker(Self);
end;

destructor TBufferMarker.Destroy;
begin
	Clear;
	FBuffer.RemoveMarker(Self);
	inherited;
end;

function TBufferMarker.FindNext(Position: Natural): PMark;
begin
	Result := FMark;
	while Result <> nil do
	begin
		if Result^.Last > Position then Break;
		Result := Result^.Next;
	end;
end;

function TBufferMarker.FindPrior(Position: Natural): PMark;
var
	Next: PMark;
begin
	Result := nil;
	Next := FMark;
	while Next <> nil do
	begin
		if Next^.First >= Position then Break;
		Result := Next;
		Next := Next^.Next;
	end;
end;

procedure TBufferMarker.Mark(First, Last: Natural);
var
	P, Pre, New, Next: PMark;
begin
	Pre := nil;
	P := FMark;

	while P <> nil do
	begin
		if P^.First > Last then
		begin
			GetMem(New, SizeOf(TMark));
			New^.Next := P;
			New^.First := First;
			New^.Last := Last;
			if Pre <> nil then Pre^.Next := New else FMark := New;
			Exit;
		end
		else if P^.Last >= First then
		begin
			if First < P^.First then P^.First := First;
			if Last > P^.Last then
			begin
				P^.Last := Last;

				Pre := P;
				P := P^.Next;
				while (P <> nil) and (P^.First <= Last) do
				begin
					if P^.Last > Pre^.Last then Pre^.Last := P^.Last;
					Next := P^.Next;
					FreeMem(P);
					P := Next;
				end;
				Pre^.Next := P;
			end;
			Exit;
		end;

		Pre := P;
		P := P^.Next;
	end;

	GetMem(P, SizeOf(TMark));
	P^.Next := nil;
	P^.First := First;
	P^.Last := Last;

	if Pre = nil then
	begin
		Assert(FMark = nil);
		FMark := P;
	end
	else
	begin
		Assert(Pre^.Next = nil);
		Pre^.Next := P;
	end;
end;

function TBufferMarker.Marked: Boolean;
begin
	Result := FMark <> nil
end;

function TBufferMarker.MarkedRange(First, Last: Natural; var Pinpoint: Boolean): Boolean;
var
	P: PMark;
begin
	P := FMark;
	while P <> nil do
	begin
		{if (P^.First < Last) and (First < P^.Last) or
			(P^.First = P^.Last) and (First = P^.First) then}
		if (P^.First >= First) and (P^.First < Last) or
			(First >= P^.First) and (First < P^.Last) then
		begin
			Pinpoint := P^.First = P^.Last; 
			Result := True;
			Exit;
		end;
		P := P^.Next;
	end;
	Result := False;
end;

{ TBuffer }

procedure TBuffer.AddMarker(AMarker: TBufferMarker);
begin
	SetLength(FMarkers, Length(FMarkers) + 1);
	FMarkers[High(FMarkers)] := AMarker;
end;

procedure TBuffer.AfterChange(First, OldLast, NewLast: Natural);
var
	I: Integer;
begin
	inherited;
	{Update Markers}
	for I := Low(FMarkers) to High(FMarkers) do
		FMarkers[I].BufferChanged(First, OldLast, NewLast);
	{Modified Mark}
	FModified.Mark(First, NewLast);
end;

function TBuffer.AllocateSubBuffer: TSubBufferRange;
var
	I: TSubBufferRange;
	S: Integer;
begin
	for I := Low(TSubBufferRange) to High(TSubBufferRange) do
	begin
		if FSub[I] = nil then
		begin
			S := Capacity;
			if S = 0 then S := 1;
			GetMem(FSub[I], S);
			Result := I;
			Exit;
		end;
	end;
	Assert(False);
	Result := 0;
end;

function TBuffer.CanRedo: Boolean;
begin
	Result := FRedo <> nil
end;

function TBuffer.CanUndo: Boolean;
begin
	Result := FUndo <> nil
end;

procedure TBuffer.CheckReadOnly;
begin
	if FReadOnly then
		raise EBufferReadOnly.Create('');
end;

procedure TBuffer.ClearModified;
begin
	FModified.Clear
end;

procedure TBuffer.ClearRedoBuffer;
var
	R: PUndoBuffer;
begin
	while FRedo <> nil do
	begin
		R := FRedo;
		FRedo := FRedo.Next;
		FreeMem(R);
	end;
end;

procedure TBuffer.ClearUndoBuffer;
var
	U: PUndoBuffer;
begin
	while FUndo <> nil do
	begin
		U := FUndo;
		FUndo := FUndo.Next;
		FreeMem(U);
	end;
	ClearRedoBuffer
end;

constructor TBuffer.Create;
begin
	inherited;
	FModified := TBufferMarker.Create(Self);
end;

destructor TBuffer.Destroy;
begin
	FModified.Free;
	inherited;
end;

function TBuffer.GetSub(ASub: TSubBufferRange; Position: Natural): Byte;
begin
	Assert(Position < Size);
	if Position < GapFirst then
		Result := PByte(Cardinal(FSub[ASub]) + Position)^
	else
		Result := PByte(Cardinal(FSub[ASub]) + Position + Cardinal(GapLast - GapFirst))^
end;

procedure TBuffer.InternalClear;
var
	I: TSubBufferRange;
begin
	for I := Low(TSubBufferRange) to High(TSubBufferRange) do
	begin
		if FSub[I] <> nil then
		begin
			FreeMem(FSub[I]);
      GetMem(FSub[I], 1);
		end
	end;
	inherited;
	ClearModified;
	ClearUndoBuffer;
end;

procedure TBuffer.MakeUndoBuffer(var Result: PUndoBuffer; Action: PUndoBuffer);
var
	P: Pointer;
begin
	GetMem(Result, SizeOf(TUndoBuffer) + Abs(Action^.DeleteLength));
	Result^.Position := Action^.Position;
	if Action^.InsertLength < 0 then Dec(Result^.Position, Action^.InsertLength);
	if Action^.DeleteLength < 0 then Inc(Result^.Position, Action^.DeleteLength);
	Result^.DeleteLength := Action^.InsertLength;
	Result^.InsertLength := Action^.DeleteLength;
	P := Pointer(Cardinal(Result) + SizeOf(TUndoBuffer));
	if Action^.DeleteLength > 0 then
		Read(Action^.Position, Action^.DeleteLength, P^)
	else
		Read(Action^.Position + Action^.DeleteLength, - Action^.DeleteLength, P^);
end;

function TBuffer.Modified: Boolean;
begin
	Result := FModified.Marked
end;

function TBuffer.ModifiedRange(First, Last: Natural): Boolean;
var
	Pinpoint: Boolean;
begin
	Result := FModified.MarkedRange(First, Last, Pinpoint)
end;

procedure TBuffer.MoveBlock(Source, Dest, Size: Natural);
var
	I: TSubBufferRange;
begin
	inherited;
	for I := High(FSub) downto Low(FSub) do
		if FSub[I] <> nil then
			System.Move(
				Pointer(Cardinal(FSub[I]) + Source)^,
				Pointer(Cardinal(FSub[I]) + Dest)^, Size);
end;

procedure TBuffer.ReallocateBlock(NewCapacity,
	From1, To1, Size1, From2, To2,	Size2, From3, To3, Size3: Natural);
var
	S: Integer;
	NewData: Pointer;
begin
	inherited;
	for S := High(FSub) downto Low(FSub) do
		if FSub[S] <> nil then
		begin
			GetMem(NewData, NewCapacity);
			System.Move(Pointer(Cardinal(FSub[S]) + From1)^, Pointer(Cardinal(NewData) + To1)^, Size1);
			System.Move(Pointer(Cardinal(FSub[S]) + From2)^, Pointer(Cardinal(NewData) + To2)^, Size2);
			System.Move(Pointer(Cardinal(FSub[S]) + From3)^, Pointer(Cardinal(NewData) + To3)^, Size3);
			if FSub[S] <> nil then FreeMem(FSub[S]);
			FSub[S] := NewData;
		end;
end;

function TBuffer.Redo: Natural;
var
	CurrentRedo, NewUndo: PUndoBuffer;
	P: Pointer;
begin
  CheckReadOnly;
	CurrentRedo := FRedo;
	FRedo := CurrentRedo^.Next;

	MakeUndoBuffer(NewUndo, CurrentRedo);
	NewUndo^.Next := FUndo;
	FUndo := NewUndo;

	Result := NewUndo^.Position;

	P := Pointer(Cardinal(CurrentRedo) + SizeOf(TUndoBuffer));
	Edit(CurrentRedo^.Position, CurrentRedo^.DeleteLength, CurrentRedo^.InsertLength, P^);

	FreeMem(CurrentRedo);
end;

function TBuffer.RedoPosition: Natural;
begin
	Result := FRedo^.Position
end;

procedure TBuffer.ReleaseSubBuffer(AIndex: TSubBufferRange);
begin
	Assert(FSub[AIndex] <> nil);
	FreeMem(FSub[AIndex]);
	FSub[AIndex] := nil;
end;

procedure TBuffer.RemoveMarker(AMarker: TBufferMarker);
var
	I, J: Integer;
begin
	for I := High(FMarkers) downto Low(FMarkers) do
		if FMarkers[I] = AMarker then
		begin
			for J := I to High(FMarkers) - 1 do
				FMarkers[J] := FMarkers[J + 1];
			SetLength(FMarkers, Length(FMarkers) - 1);
			Break;
		end;
end;

procedure TBuffer.Replace(Position: Natural; DeleteLength, InsertLength: Integer; const Data);
var
	Undo: PUndoBuffer;
	P: Pointer;
begin
	CheckReadOnly;
	ClearRedoBuffer;
	GetMem(Undo, SizeOf(TUndoBuffer) + Abs(DeleteLength));

	Undo.Position := Position;
	if InsertLength < 0 then Dec(Undo.Position, InsertLength);
	if DeleteLength < 0 then Inc(Undo.Position, DeleteLength);
	Undo.DeleteLength := InsertLength;
	Undo.InsertLength := DeleteLength;
	P := Pointer(Cardinal(Undo) + SizeOf(TUndoBuffer));
	if DeleteLength > 0 then
		Read(Position, DeleteLength, P^)
	else
		Read(Position + DeleteLength, - DeleteLength, P^);

	Undo^.Next := FUndo;
	FUndo := Undo;

	inherited;
end;

procedure TBuffer.SetSub(ASub: TSubBufferRange; Position: Natural; const Value: Byte);
begin
	Assert(Position < Size);
	if Position < GapFirst then
		PByte(Cardinal(FSub[ASub]) + Position)^ := Value
	else
		PByte(Cardinal(FSub[ASub]) + Position + Cardinal(GapLast - GapFirst))^ := Value
end;

function TBuffer.Undo: Natural;
var
	CurrentUndo, NewRedo: PUndoBuffer;
	P: Pointer;
begin
  CheckReadOnly;
	CurrentUndo := FUndo;
	FUndo := CurrentUndo^.Next;

	MakeUndoBuffer(NewRedo, CurrentUndo);
	NewRedo^.Next := FRedo;
	FRedo := NewRedo;

	Result := NewRedo^.Position;

	P := Pointer(Cardinal(CurrentUndo) + SizeOf(TUndoBuffer));
	Edit(CurrentUndo^.Position, CurrentUndo^.DeleteLength, CurrentUndo^.InsertLength, P^);

	FreeMem(CurrentUndo);
end;

function TBuffer.UndoPosition: Natural;
begin
	Result := FUndo^.Position
end;

end.

