unit BinaryDecoder;

interface

uses
	Windows, SysUtils,
	BinaryBuffer;

const
	UnicodeLineFeed = WideChar($2029);
	UnicodeFormFeed = WideChar($0c);
	UnicodeFullWidthSpace = WideChar($3000);
	UnicodeBOM = WideChar($feff);

	LineFeedCodes: WideString = UnicodeLineFeed + WideString(#13 + #10) + UnicodeFormFeed;

type
	TLineFeedCode = (lfNewLine, lfNewLineCR, lfNewLineLF, lfFormFeed);

	{R[h}
	TLetterBinaryCode = packed record
		Length: Cardinal;
		Codes: array[0..3] of Byte;
	end;

	TDecoder = class of AbstractDecoder;

	AbstractDecoder = class(TObject)
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; virtual; abstract;
		class function Encode(Letter: WideChar): TLetterBinaryCode; virtual; abstract;
		class function EncodeString(const S: WideString): string;
		class function LineFeed(const Accessor: TBufferAccessor): Boolean; virtual;
	end;

	AbstractMBCSDecoder = class(AbstractDecoder)
	public
		class function LineFeed(const Accessor: TBufferAccessor): Boolean; override;
	end;

	AsciiDecoder = class(AbstractMBCSDecoder)
	public const
		ENGLISH = 1252;
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; override;
		class function Encode(Letter: WideChar): TLetterBinaryCode; override;
	end;

	LocaleDecoder = class(AbstractMBCSDecoder)
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; override;
		class function Encode(Letter: WideChar): TLetterBinaryCode; override;
	end;

	UTF8Decoder = class(AbstractMBCSDecoder)
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; override;
		class function Encode(Letter: WideChar): TLetterBinaryCode; override;
	end;

	UTF16Decoder = class(AbstractDecoder)
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; override;
		class function Encode(Letter: WideChar): TLetterBinaryCode; override;
	end;

	UTF16BEDecoder = class(AbstractDecoder)
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; override;
		class function Encode(Letter: WideChar): TLetterBinaryCode; override;
	end;

	EUCJPDecoder = class(AbstractMBCSDecoder)
	public const
		EUCJP = 20932; {54932}
		EUCJPLeadBytes = [#$8e, #$a1..#$fe];
	public
		class function Decode(var Accessor: TBufferAccessor): WideChar; override;
		class function Encode(Letter: WideChar): TLetterBinaryCode; override;
	end;

function SelectDecoder(S: PChar; L: Integer): TDecoder;
function SelectLineFeed(Decoder: TDecoder; S: PChar; L: Integer): TLineFeedCode;

const
	LineFeedCode: array[TLineFeedCode] of WideChar =
		(UnicodeLineFeed, #13, #10, UnicodeFormFeed);

implementation

(*
function ShiftJisToJis(Code: Word): Word; register;
asm
	and ah, 3Fh
	shl ah, 1
	sub al, 9Fh
	jae @SKIP
	dec ah
	cmp al, 80h-9Fh
	adc al, 9Fh-40h-1
@SKIP:
	add ax, 2021h
end;

function JisToShiftJis(Code: Word): Word; register;
asm
	add ax, (9Fh-21h) + (42h-21h) * 256
	shr ah, 1
	jc @SKIP
	cmp al, 60h+(9Fh-21h)
	sbb al, 9Fh-40h-1
@SKIP:
	xor ah, 0A0h
end;

function EucJPToJis(Code: Word): Word; register;
asm
	sub ax, 8080h
end;

function JisToEucJP(Code: Word): Word; register;
asm
	add ax, 8080h
end;
*)

function SelectDecoder(S: PChar; L: Integer): TDecoder;
label
	UTF16, UTF8, EUC, Locale, Ascii;
const
	EUC_NO = $cea4;
var
	I, LE, BE: Integer;
	W: PWideChar;
	F, H: Boolean;
begin
	if L = 0 then
	begin
		Result := LocaleDecoder;
		Exit;
	end;

UTF16:
	W := PWideChar(S);

	if W^ = #$feff then
	begin
		Result := UTF16Decoder;
		Exit;
	end;
	if W^ = #$fffe then
	begin
		Result := UTF16BEDecoder;
		Exit;
	end;

	LE := 0;
	BE := 0;
	for I := 0 to (L div 2) - 1 do
	begin
		if ((W + I)^ = #$000d) or ((W + I)^ = #$000a) then
		begin
			Inc(LE);
			if LE >= 4 then
			begin
				Result := UTF16Decoder;
				Exit;
			end
		end;
		if ((W + I)^ = #$0d00) or ((W + I)^ = #$0a00) then
		begin
			Inc(BE);
			if BE >= 4 then
			begin
				Result := UTF16BEDecoder;
				Exit;
			end
		end;
	end;

UTF8:
	//UTF-8ƌĈᔽUTF-8

	if (L >= 3) and (PCardinal(S)^ and $00ffffff <> $bfbbef) then
	begin
		F := False;
		I := 0;
		while I < L do
		begin
			case (S + I)^ of
			#0..#$7f:
				Inc(I);
			#$c0..#$df:
			begin
				if (I + 1 < L) and not ((S + I + 1)^ in [#$80..#$bf]) then goto EUC;
				Inc(I, 2);
				F := True;
			end;
			#$e0..#$ef:
			begin
				if (I + 1 < L) and not ((S + I + 1)^ in [#$80..#$bf]) then goto EUC;
				if (I + 2 < L) and not ((S + I + 2)^ in [#$80..#$bf]) then goto EUC;
				Inc(I, 3);
				F := True;
			end;
			else
				goto EUC;
			end;
		end;
		if not F then goto EUC;
	end;
	Result := UTF8Decoder;
	Exit;

EUC:
	//EUC-JPƌĖúv݂EUC-JP
	F := False;
	I := 0;
	while I < L do
	begin
		case (S + I)^ of
		#$8e, #$a1..#$fe:
		begin
			if (I + 1 < L) and not ((S + I + 1)^ in [#$80..#$fe]) then goto Locale;
			if PWord(S + I)^ = EUC_NO then F := True;
			Inc(I, 2);
		end;
		else
			Inc(I);
		end;
	end;
	if not F then goto Locale;
	Result := EUCJPDecoder;
	Exit;

Locale:
	//IsDBCSLeadByte̎#$40..#$fcȂLocale
	//pJ^Ji݂SpAscii
	F := False;
	H := False;
	I := 0;
	while I < L do
	begin
		if IsDBCSLeadByte(Ord((S + I)^)) then
		begin
			if (I + 1 < L) and not ((S + I + 1)^ in [#$40..#$fc]) then goto Ascii;
			Inc(I, 2);
			F := True;
		end
		else if (S + I)^ >= #$80 then
		begin
			Inc(I);
			H := True; {}
		end
		else
			Inc(I);
	end;
	if H and not F then goto Ascii;
	Result := LocaleDecoder;
	Exit;

Ascii:
	Result := AsciiDecoder;
end;

function SelectLineFeed(Decoder: TDecoder; S: PChar; L: Integer): TLineFeedCode;
var
	W: PWideChar;
begin
	Result := lfNewLine;
	if Decoder = UTF16Decoder then
	begin
		W := PWideChar(S);
		L := L div 2;
		while L > 0 do
		begin
			if (W^ = #$000d) and (L > 1) and ((W + 1)^ = #$000a) then
				Break
			else if W^ = #$000d then
			begin
				Result := lfNewLineCR;
				Break;
			end
			else if W^ = #$000a then
			begin
				Result := lfNewLineLF;
				Break;
			end;
			Inc(W);
			Dec(L);
		end;
	end
	else if Decoder = UTF16BEDecoder then
	begin
		W := PWideChar(S);
		L := L div 2;
		while L > 0 do
		begin
			if (W^ = #$0d00) and (L > 1) and ((W + 1)^ = #$0a00) then
				Break
			else if W^ = #$0d00 then
			begin
				Result := lfNewLineCR;
				Break;
			end
			else if W^ = #$0a00 then
			begin
				Result := lfNewLineLF;
				Break;
			end;
			Inc(W);
			Dec(L);
		end;
	end
	else
	begin
		while L > 0 do
		begin
			if (S^ = #13) and (L > 1) and ((S + 1)^ = #10) then
				Break
			else if S^ = #13 then
			begin
				Result := lfNewLineCR;
				Break;
			end
			else if S^ = #10 then
			begin
				Result := lfNewLineLF;
				Break;
			end;
			Inc(S);
			Dec(L);
		end;
	end;
end;

{ AbstractDecoder }

class function AbstractDecoder.EncodeString(const S: WideString): string;
var
	I: Integer;
	R: TLetterBinaryCode;
	L: Cardinal;
	P: PChar;
begin
	Result := '';
	for I := 1 to Length(S) do
	begin
		R := Encode(S[I]);
		L := Length(Result);
		SetLength(Result, L + R.Length);
		P := Pointer(Cardinal(Result) + L);
		case R.Length of
		1:
			PByte(P)^ := R.Codes[0];
		2:
			PWORD(P)^ := PWORD(@R.Codes[0])^;
		3:
		begin
			PByte(P)^ := R.Codes[0];
			PWORD(P + 1)^ := PWORD(@R.Codes[1])^;
		end;
		else
			PDWORD(P)^ := PDWORD(@R.Codes[0])^;
		end;
	end;
end;

class function AbstractDecoder.LineFeed(const Accessor: TBufferAccessor): Boolean;
var
	A2: TBufferAccessor;
begin
	if Accessor.EndOfBuffer then
		Result := False
	else
	begin
		A2 := Accessor;
		case Decode(A2) of
		#10, #13, UnicodeLineFeed, UnicodeFormFeed:
			Result := True;
		else
			Result := False;
		end;
	end	
end;

{ AbstractMBCSDecoder }

class function AbstractMBCSDecoder.LineFeed(const Accessor: TBufferAccessor): Boolean;
begin
	Result := not Accessor.EndOfBuffer and
		(Accessor.Buffer[Accessor.Index] in [10, 12, 13]);
end;

{ AsciiDecoder }

class function AsciiDecoder.Decode(var Accessor: TBufferAccessor): WideChar;
var
	C: Char;
begin
	C := Chr(Accessor.Take);
	if C = #13 then
	begin
		Result := #13;
		if Accessor.Match(10) then
		begin
			Accessor.Advance;
			Result := UnicodeLineFeed
		end;
	end
	else
	begin
		MultiByteToWideChar(English, MB_USEGLYPHCHARS, @C, 1, @Result, 1);
	end;
end;

class function AsciiDecoder.Encode(Letter: WideChar): TLetterBinaryCode;
var
	Buffer: array[0..7] of Char;
	R: Integer;
begin
	if Letter = UnicodeLineFeed then
	begin
		Result.Length := 2;
		Result.Codes[0] := 13;
		Result.Codes[1] := 10;
	end
	else
	begin
		R := WideCharToMultiByte(1252, 0, @Letter, 1, Buffer, Length(Result.Codes), nil, nil);
		if R = 1 then
		begin
			Result.Length := 1;
			Result.Codes[0] := Ord(Buffer[0]);
		end
		else
		begin
			Result.Length := 1;
			Result.Codes[0] := Ord('?');
		end;
	end
end;

{ LocaleDecoder }

class function LocaleDecoder.Decode(var Accessor: TBufferAccessor): WideChar;
var
	C: Char;
	Cs: array[0..1] of Char;
begin
	C := Char(Accessor.Take);
	if C < #128 then
	begin
		Result := WideChar(C);
		if (C = #13) and Accessor.Match(10) then
		begin
			Accessor.Advance;
			Result := UnicodeLineFeed;
		end;
	end
	else if C in LeadBytes then
	begin
		if Accessor.EndOfBuffer then
			Result := WideChar(C)
		else
		begin
			Cs[0] := C;
			Cs[1] := Chr(Accessor.Take);
			MultiByteToWideChar(CP_ACP, MB_USEGLYPHCHARS, Cs, 2, @Result, 1);
		end
	end
	else
	begin
		MultiByteToWideChar(CP_ACP, MB_USEGLYPHCHARS, @C, 1, @Result, 1);
	end;
end;

class function LocaleDecoder.Encode(Letter: WideChar): TLetterBinaryCode;
var
	Buffer: array[0..7] of Char;
	R: Integer;
begin
	if Letter = UnicodeLineFeed then
	begin
		Result.Length := 2;
		Result.Codes[0] := 13;
		Result.Codes[1] := 10;
	end
	else
	begin
		R := WideCharToMultiByte(CP_ACP, 0, @Letter, 1, Buffer, Length(Result.Codes), nil, nil);
		if R <= 2 then
		begin
			Result.Length := R;
			Result.Codes[0] := Ord(Buffer[0]);
			Result.Codes[1] := Ord(Buffer[1]);
		end
		else
		begin
			Result.Length := 1;
			Result.Codes[0] := Ord('?');
		end;
	end
end;

{ UTF8Decoder }

class function UTF8Decoder.Decode(var Accessor: TBufferAccessor): WideChar;
var
	R, C: Word;
begin
	R := Accessor.Take;
	if R and $80 <> 0 then
	begin
		R := R and $3f;
		if R and $20 <> 0 then
		begin
			if Accessor.EndOfBuffer then begin Result := '?'; Exit; end;
			C := Accessor.Take;
			R := R shl 6 or C and $3f;
		end;
		if Accessor.EndOfBuffer then begin Result := '?'; Exit; end;
		C := Accessor.Take;
		R := R shl 6 or C and $3f;
	end
	else if R = 13 then
	begin
		if Accessor.Match(10) then
		begin
			Accessor.Advance;
			R := Word(UnicodeLineFeed);
		end;
	end;
	Result := WideChar(R);
end;

class function UTF8Decoder.Encode(Letter: WideChar): TLetterBinaryCode;
var
	Codes: array[0..7] of Char;
begin
	if Letter = UnicodeLineFeed then
	begin
		Result.Length := 2;
		Result.Codes[0] := 13;
		Result.Codes[1] := 10;
	end
	else
	begin
		Result.Length := UnicodeToUtf8(Codes, 7, @Letter, 1) - 1;
		if Result.Length > 3 then Result.Length := 3;
		Result.Codes[0] := Ord(Codes[0]);
		Result.Codes[1] := Ord(Codes[1]);
		Result.Codes[2] := Ord(Codes[2]);
	end
end;

{ UTF16Decoder }

class function UTF16Decoder.Decode(var Accessor: TBufferAccessor): WideChar;
begin
	WordRec(Result).Lo := Accessor.Take;
	if Accessor.EndOfBuffer then
		WordRec(Result).Hi := 0
	else
	begin
		WordRec(Result).Hi := Accessor.Take;
		if (Result = #13) and Accessor.Match2(10, 0) then
		begin
			Accessor.Advance2;
			Result := UnicodeLineFeed
		end;
	end;
end;

class function UTF16Decoder.Encode(Letter: WideChar): TLetterBinaryCode;
begin
	if Letter = UnicodeLineFeed then
	begin
		Result.Length := 4;
		Result.Codes[0] := 13;
		Result.Codes[1] := 0;
		Result.Codes[2] := 10;
		Result.Codes[3] := 0;
	end
	else
	begin
		Result.Length := 2;
		Result.Codes[0] := Lo(Word(Letter));
		Result.Codes[1] := Hi(Word(Letter));
	end
end;

{ UTF16BEDecoder }

class function UTF16BEDecoder.Decode(var Accessor: TBufferAccessor): WideChar;
begin
	WordRec(Result).Hi := Accessor.Take;
	if Accessor.EndOfBuffer then
		WordRec(Result).Lo := 0
	else
	begin
		WordRec(Result).Lo := Accessor.Take;
		if (Result = #13) and Accessor.Match2(0, 10) then
		begin
			Accessor.Advance2;
			Result := UnicodeLineFeed
		end;
	end;
end;

class function UTF16BEDecoder.Encode(Letter: WideChar): TLetterBinaryCode;
begin
	if Letter = UnicodeLineFeed then
	begin
		Result.Length := 4;
		Result.Codes[0] := 0;
		Result.Codes[1] := 13;
		Result.Codes[2] := 0;
		Result.Codes[3] := 10;
	end
	else
	begin
		Result.Length := 2;
		Result.Codes[0] := Hi(Word(Letter));
		Result.Codes[1] := Lo(Word(Letter));
	end
end;

{ EUCJPDecoder }

class function EUCJPDecoder.Decode(var Accessor: TBufferAccessor): WideChar;
var
	C: Char;
	Cs: array[0..1] of Char;
begin
	C := Char(Accessor.Take);
	if C < #128 then
	begin
		Result := WideChar(C);
		if (C = #13) and Accessor.Match(10) then
		begin
			Accessor.Advance;
			Result := UnicodeLineFeed;
		end;
	end
	else if C in EUCJPLeadBytes then
	begin
		if Accessor.EndOfBuffer then
			Result := WideChar(C)
		else
		begin
			Cs[0] := C;
			Cs[1] := Chr(Accessor.Take);
			MultiByteToWideChar(EUCJP, MB_USEGLYPHCHARS, Cs, 2, @Result, 1);
		end
	end
	else
	begin
		MultiByteToWideChar(EUCJP, MB_USEGLYPHCHARS, @C, 1, @Result, 1);
	end;
end;

class function EUCJPDecoder.Encode(Letter: WideChar): TLetterBinaryCode;
var
	Buffer: array[0..7] of Char;
	R: Integer;
begin
	if Letter = UnicodeLineFeed then
	begin
		Result.Length := 2;
		Result.Codes[0] := 13;
		Result.Codes[1] := 10;
	end
	else
	begin
		R := WideCharToMultiByte(EUCJP, 0, @Letter, 1, Buffer, Length(Result.Codes), nil, nil);
		if R <= 2 then
		begin
			Result.Length := R;
			Result.Codes[0] := Ord(Buffer[0]);
			Result.Codes[1] := Ord(Buffer[1]);
		end
		else
		begin
			Result.Length := 1;
			Result.Codes[0] := Ord('?');
		end;
	end
end;

end.

