unit BinaryEditor;

interface

uses
	SysUtils, Types,
	BinaryBuffer, BinaryDecoder;

const
	SubStateMask     = $1f;
	//SubStateTrailing = SubStateMask;

	SubStateBase = $20;
	SubLineEnd   = $40;
	SubLetter    = $80;

type
	TState = 0..$1f;

	TLine = -1..High(Integer);
  TLines = array of Natural;

	TLayout = class;
	TView = class;
	IViewUI = interface;

	TScrollBase = (sbTop, sbBottom);

	{oCif[^}
	TDocument = class(TBuffer)
	private
		FLayouts: array of TLayout;
	protected
		procedure AddLayout(const Layout: TLayout);
		procedure RemoveLayout(const Layout: TLayout);
		procedure BeforeChange(First: Natural; Last: Natural); override;
		procedure AfterChange(First: Natural; OldLast: Natural; NewLast: Natural); override;
	public
		procedure Encode(Decoder, Encoder: TDecoder);
	end;

	{}
	TLayout = class(TObject)
	private
		FDocument: TDocument;
		FSub: TSubBufferRange;
		FTotalLineCount: Integer;
		FViews: array of TView;
		FDecoder: TDecoder;
		FLineWidth: Integer;
		FLineHeight: Integer;
		FTabWidth: Integer;
		FMaxWidth: Integer;
		FHanging: WideString; {"." "!"ȂǂԂ牺}
		FExpelling: WideString; {"(" "[" Ȃǂǂo}
		FIndenting: WideString; {^u"*"ȂǂŎn܂s̐܂Ԃ̓Vtg}
		FMaxIndentWidth: Integer;
		FFixedLetterWidth: Integer;
		FWordWrap: Boolean;
		FLineModified: Boolean;
		function GetSub(AIndex: Natural): Byte;
		procedure SetSub(AIndex: Natural; const Value: Byte);
	protected
		procedure AddView(const View: TView);
		procedure RemoveView(const View: TView);
		procedure BeforeChange(First: Natural; Last: Natural);
		procedure AfterChange(First: Natural; OldLast: Natural; NewLast: Natural);
		function GetTabWidth(X: Integer): Integer;
		function IsLogicalLineTop(Position: Natural): Boolean;
		function GetLogicalLineTop(Position: Natural): TLine;
		function GetLogicalLineEnd(Position: Natural): TLine;
		function GetLogicalLineFromLineNumber(LineNumber: Integer): TLine;
		function LogicalPredLine(Line: TLine): TLine;
		function LogicalSuccLine(Line: TLine): TLine;
		function PredLine(Line: TLine): TLine;
		function SuccLine(Line: TLine): TLine;
		function SuccLetter(Position: Natural): Natural;
		function PredLetter(Position: Natural): Natural;
		function LetterTop(Position: Natural): Natural;
		function LetterCount(Position: Natural): Natural;
	public
		constructor Create(
			ADocument: TDocument;
			ADecoder: TDecoder;
			ALineWidth: Integer;
			ATabWidth: Integer;
			AFixedLetterWidth: Integer;
			ALineHeight: Integer;
			AHanging: WideString;
			AExpelling: WideString;
			AIndenting: WideString;
			AWordWrap: Boolean);
		destructor Destroy; override;
		function LetterWidth(Letter: WideChar): Integer; virtual; abstract;
		function StandardLetterWidth: Integer; virtual; abstract;
		procedure Analyze(var First, Last: Natural); virtual;
		function ViewCount: Integer;
		function GetLines(IncludingEnd: Boolean): TLines;
		property TotalLineCount: Integer read FTotalLineCount;
		property LineWidth: Integer read FLineWidth;
		property TabWidth: Integer read FTabWidth;
		property FixedLetterWidth: Integer read FFixedLetterWidth;
		property Hanging: WideString read FHanging;
		property Expelling: WideString read FExpelling;
		property Indenting: WideString read FIndenting;
		property MaxIndentWidth: Integer read FMaxIndentWidth;
		property LineHeight: Integer read FLineHeight;
		property WordWrap: Boolean read FWordWrap;
		property Document: TDocument read FDocument;
		property Decoder: TDecoder read FDecoder;
		property SubIndex: TSubBufferRange read FSub;
		property Sub[AIndex: Natural]: Byte read GetSub write SetSub;
	end;

	{\C}
	TViewClass = class of TView;
	TView = class(TObject)
	private
		FLayout: TLayout;
		FViewUI: IViewUI;
	protected
		procedure BeforeChange(First: Natural; Last: Natural); virtual;
		procedure AfterChange(First, OldLast, NewLast, RepaintingFirst, RepaintingLast: Natural; LineModified: Boolean); virtual; abstract;
		procedure VertScrollBarInternal(LineCount, LinePosition: Integer; out Position, Min, Max, PageSize: Integer);
	public
		constructor Create(ALayout: TLayout; const AViewUI: IViewUI); virtual;
		destructor Destroy; override;
		property Layout: TLayout read FLayout;
		function ViewType: TViewClass;
		function ReadSingleLetter(Position: Natural): TLetterBinaryCode;
		function PredLine(Position: Natural): TLine; virtual; abstract;
		function SuccLine(Position: Natural): TLine; virtual; abstract;
		function LineEnd(Position: Natural): Natural;
		function ToLine(Position: Natural; Dir: Integer): TLine; virtual; abstract;
		function ToX(Line: TLine; Position: Natural): Integer; virtual; abstract;
		function ToPosition(Line: TLine; X: Integer): Natural; virtual; abstract;
		function InsertData(Position: Natural; DeleteLength: Integer; const Data: string; HexMode: Boolean): Natural;
		function InsertText(Position: Natural; DeleteLength: Integer; const Data: WideString; HexMode: Boolean): Natural;
		function ReadData(First, Last: Natural): string;
		function ReadText(First, Last: Natural): WideString;
		function FixCaretPosition(Position: Natural): Natural;
		function LeftCaretPosition(HexMode: Boolean; Position: Natural): Natural; virtual; abstract;
		function RightCaretPosition(HexMode: Boolean; Position: Natural): Natural; virtual; abstract;
		function UpCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural; virtual;
		function DownCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural; virtual;
		function LineTopCaretPosition(Position: Natural; Smart: Boolean = False): Natural; virtual; abstract;
		function LineEndCaretPosition(Position: Natural): Natural; virtual; abstract;
		function DeleteSingleLetter(Position: Natural; HexMode: Boolean; Dir: Integer): Natural;
		procedure ReadyPaintLine(Base: TScrollBase; From: Natural; out Y: Integer; out NewLine: Boolean; out LineNumber, Indent: Integer; out Accessor: TBufferAccessor); virtual; abstract;
		procedure HorzScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer); virtual; abstract;
		procedure VertScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer); virtual; abstract;
		procedure VertScroll(var Base: TScrollBase; var From: Natural; Position, Min, Max, PageSize: Integer);
		function ScrollUp(var Base: TScrollBase; var From: Natural): Integer;
		function ScrollDown(var Base: TScrollBase; var From: Natural): Integer;
		function BasedScrollUp(var Base: TScrollBase; var From: Natural): Integer;
		function BasedScrollDown(var Base: TScrollBase; var From: Natural): Integer;
		procedure AdjustScroll(var Left: Integer; var Base: TScrollBase; var From: Natural; out ScrollX, ScrollY: Integer);
		procedure AdjustRepaint(Base: TScrollBase; From: Natural; First, Last: Natural; out Y1, Y2: Integer);
	end;

	TTextView = class(TView)
	private
		FLineNumberFrom: TLine;
		FLineNumber: Integer;
		FScreenNumber: Integer;
	protected
		procedure BeforeChange(First: Natural; Last: Natural); override;
		procedure AfterChange(First, OldLast, NewLast, RepaintingFirst, RepaintingLast: Natural; LineModified: Boolean); override;
	public
		constructor Create(ALayout: TLayout; const AViewUI: IViewUI); override;
		function PredLine(Position: Natural): TLine; override;
		function SuccLine(Position: Natural): TLine; override;
		function ToLine(Position: Natural; Dir: Integer): TLine; override;
		function LeftCaretPosition(HexMode: Boolean; Position: Natural): Natural; override;
		function RightCaretPosition(HexMode: Boolean; Position: Natural): Natural; override;
		function ToX(Line: TLine; Position: Natural): Integer; override;
		function ToPosition(Line: TLine; X: Integer): Natural; override;
		function LineTopCaretPosition(Position: Natural; Smart: Boolean = False): Natural; override;
		function LineEndCaretPosition(Position: Natural): Natural; override;
		procedure ReadyPaintLine(Base: TScrollBase; From: Natural; out Y: Integer; out NewLine: Boolean; out LineNumber, Indent: Integer; out Accessor: TBufferAccessor); override;
		procedure HorzScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer); override;
		procedure VertScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer); override;
		function TabWidth(X: Integer): Integer;
		function IndentWidth(Line: TLine): Integer;
		function AutoIndent(Position: Natural): WideString;
		function Indent(First, Last: Natural): Natural;
		function Unindent(First, Last: Natural): Natural;
		procedure GetLocation(Position: Natural; out ScreenNumber, ScreenColumn, LineNumber, Column: Integer);
		procedure AdjustScrollOnChange(First, OldLast, NewLast: Natural; var Base: TScrollBase; var From: Natural);
	end;

	TBinaryView = class(TView)
	protected
		procedure AfterChange(First, OldLast, NewLast, RepaintingFirst, RepaintingLast: Natural; LineModified: Boolean); override;
	public
		function PredLine(Position: Natural): TLine; override;
		function SuccLine(Position: Natural): TLine; override;
		function ToLine(Position: Natural; Dir: Integer): TLine; override;
		function LeftCaretPosition(HexMode: Boolean; Position: Natural): Natural; override;
		function RightCaretPosition(HexMode: Boolean; Position: Natural): Natural; override;
		function UpCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural; override;
		function DownCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural; override;
		function ToX(Line: TLine; Position: Natural): Integer; override;
		function ToPosition(Line: TLine; X: Integer): Natural; override;
		function LineTopCaretPosition(Position: Natural; Smart: Boolean = False): Natural; override;
		function LineEndCaretPosition(Position: Natural): Natural; override;
		procedure ReadyPaintLine(Base: TScrollBase; From: Natural; out Y: Integer; out NewLine: Boolean; out LineNumber, Indent: Integer; out Accessor: TBufferAccessor); override;
		procedure HorzScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer); override;
		procedure VertScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer); override;
	end;

	{GUIGfB^}
	IViewUI = interface(IInterface)
		procedure Changed(First, OldLast, NewLast, RepaintingFirst, RepaintingLast: Natural);
		function ViewRect: TRect;
		function ScrollBase: TScrollBase;
	end;

function HasWideChar(const Str: WideString; C: WideChar): Boolean; register;
function AdjustIndent(const S, Indent, IndentLetters: WideString): WideString;

implementation

function HasWideChar(const Str: WideString; C: WideChar): Boolean; register;
asm
	test eax, eax
	jz @Exit
	push edi
	mov edi, eax
	mov ax, dx
	mov ecx, [edi - 4]
	shr ecx, 1
	repne scasw
	setz al
	pop edi
@Exit:
end;

function AdjustIndent(const S, Indent, IndentLetters: WideString): WideString;
var
	I, N, Len: Integer;
	P: PWideChar;
begin
	Result := S;

	{Ō̉s͎}

	P := Pointer(Result);
	Len := Length(Result);

	if (Len > 2) and (P[Len - 1] = #10) and (P[Len - 2] = #13) then
		Delete(Result, Len - 1, 2)
	else if Len > 1 then
	begin
		case P[Len - 1] of
		#13, #10, UnicodeLineFeed, UnicodeFormFeed:
			Delete(Result, Len, 1);
		end;
	end;

	{Text̊es̃Cfg̍ł󂢂̗̂p}
	{Text̊es炻폜Aݍs̃Cfgsڈȍ~ɒǉ}

	P := Pointer(Result);
	Len := Length(Result);

	N := 0;
	while (N < Len) and HasWideChar(IndentLetters, P[N]) do
		Inc(N);

	I := 0;
	while I < Len do
	begin
		if (P[I] = #13) and (P[I + 1] <> #10) or (P[I] = #10) or
			(P[I] = UnicodeLineFeed) or (P[I] = UnicodeFormFeed) then
		begin
			Inc(I);
			if I < Len then
			begin
				while not CompareMem(@P[I], P, N * 2) do Dec(N)
			end;
			Inc(I, N);
		end;
		Inc(I)
	end;

	Delete(Result, 1, N);
	P := PWideChar(Result);
	Len := Length(Result);

	I := 0;
	while I < Len do
	begin
		if (P[I] = #13) and (P[I + 1] <> #10) or (P[I] = #10) or
			(P[I] = UnicodeLineFeed) or (P[I] = UnicodeFormFeed) then
		begin
			Inc(I, 2);
			Delete(Result, I, N);
			Insert(Indent, Result, I);
			P := PWideChar(Result);
			Len := Length(Result);
		end;
		Inc(I)
	end;
end;

{ TDocument }

procedure TDocument.AddLayout(const Layout: TLayout);
begin
	SetLength(FLayouts, Length(FLayouts) + 1);
	FLayouts[High(FLayouts)] := Layout;
end;

procedure TDocument.AfterChange(First, OldLast, NewLast: Natural);
var
	I: Integer;
begin
	inherited;
	for I := High(FLayouts) downto Low(FLayouts) do
		FLayouts[I].AfterChange(First, OldLast, NewLast);
end;

procedure TDocument.BeforeChange(First, Last: Natural);
var
	I: Integer;
begin
	inherited;
	for I := High(FLayouts) downto Low(FLayouts) do
		FLayouts[I].BeforeChange(First, Last);
end;

procedure TDocument.Encode(Decoder, Encoder: TDecoder);
var
	Accessor: TBufferAccessor;
	Writer: TBuffer.TWriter;
	Letter: WideChar;
	Codes: TLetterBinaryCode;
begin
	Writer := TBuffer.TWriter.Create(Self);
	try
		Accessor.Buffer := Self;
		Accessor.Index := 0;
		while not Accessor.EndOfBuffer do
		begin
			Letter := Decoder.Decode(Accessor);
			Codes := Encoder.Encode(Letter);
			Writer.Write(Codes.Codes[0], Codes.Length);
		end;
	finally
		Writer.Free
	end;
end;

procedure TDocument.RemoveLayout(const Layout: TLayout);
var
	I, J: Integer;
begin
	for I := High(FLayouts) downto Low(FLayouts) do
		if FLayouts[I] = Layout then
		begin
			for J := I to High(FLayouts) - 1 do FLayouts[J] := FLayouts[J + 1];
			SetLength(FLayouts, Length(FLayouts) - 1);
			Break;
		end;
end;

{ TLayout }

procedure TLayout.AddView(const View: TView);
begin
	SetLength(FViews, Length(FViews) + 1);
	FViews[High(FViews)] := View;
end;

procedure TLayout.AfterChange(First, OldLast, NewLast: Natural);

	function IsAlphabet(Letter: WideChar): Boolean;
	begin
		Result := (Letter >= 'A') and (Letter <= 'Z') or (Letter >= 'a') and (Letter <= 'z')
	end;

	procedure MakeLine(Last: Natural);
	var
		P: Natural;
	begin
		P := Last - 1;
		Sub[P] := Sub[P] or SubLineEnd;
		Inc(FTotalLineCount);
		FLineModified := True
	end;

	function RebuildLines: Natural;
	label
		NewLine;
	var
		RebuildingFirst: Natural;
		LetterFirst, LetterLast: Natural;
		WrapPos: Natural;
		RebuildingAccessor, PreAccessor, RefAccessor: TBufferAccessor;
		Width, PreWidth, StdWidth, LineWidth, IndentWidth, ExpellingX, WrapX: Integer;
		Letter, ExpellingLetter: WideChar;
		IndentingFlag, IsNewLine: Boolean;
		I, WrapWidth: Integer;
		S: Byte;
	begin
		StdWidth := StandardLetterWidth;
		LineWidth := Self.LineWidth;

		RebuildingFirst := GetLogicalLineTop(First);
		if (RebuildingFirst > 0) and (RebuildingFirst = First) then begin
			RebuildingFirst := LogicalPredLine(RebuildingFirst);
		end;

		IsNewLine := False;
		IndentWidth := 0;
		IndentingFlag := True;
		WrapWidth := 0;
		WrapX := 0;
		WrapPos := 0;

		RebuildingAccessor.Buffer := Document;
		RebuildingAccessor.Index := RebuildingFirst;

		while not RebuildingAccessor.EndOfBuffer and (
			(RebuildingAccessor.Index <= NewLast) or
			not IsNewLine or
			(Sub[RebuildingAccessor.Index] and SubLetter = 0)) do
		begin
			Width := IndentWidth + WrapWidth;
			WrapWidth := 0;

			ExpellingX := 0;
			ExpellingLetter := #0;

			while not RebuildingAccessor.EndOfBuffer do
			begin
				PreAccessor := RebuildingAccessor;
				Letter := Decoder.Decode(RebuildingAccessor);

				PreWidth := Width;

				LetterFirst := PreAccessor.Index;
				LetterLast := RebuildingAccessor.Index;

				S := Sub[LetterFirst];

				IsNewLine := False;
				case Letter of
				#13, #10, UnicodeLineFeed, UnicodeFormFeed:
					begin
						IsNewLine := True;
						IndentWidth := 0;
						IndentingFlag := True;
						if (LetterFirst >= NewLast) and (S and SubLetter <> 0) then begin
							goto NewLine;
						end
					end;
				end;

				if (S and SubLineEnd <> 0) and ((LetterFirst < First) or (LetterFirst >= NewLast)) then begin
					Dec(FTotalLineCount);
					FLineModified := True;
				end;
				Sub[LetterFirst] := Sub[LetterFirst] and not SubLineEnd or SubLetter;

				for I := LetterFirst + 1 to LetterLast - 1 do begin
					S := Sub[I];
					if (S and SubLineEnd <> 0) and ((I < First) or (I >= NewLast)) then begin
						Dec(FTotalLineCount);
						FLineModified := True;
					end;
					Sub[I] := S and not (SubLetter or SubLineEnd);
				end;

				if IsNewLine then begin
					MakeLine(RebuildingAccessor.Index);
					goto NewLine;
				end else if Letter = #9 then begin
					Inc(Width, GetTabWidth(Width))
				end else begin
					Inc(Width, LetterWidth(Letter));
				end;

				if (LineWidth > 0) and (Width > LineWidth) then begin
					if not HasWideChar(Hanging, Letter) then begin
						RebuildingAccessor := PreAccessor;
						Width := PreWidth;
						Break
					end;
					if Width > LineWidth + StdWidth * 4 then begin
						if RebuildingAccessor.EndOfBuffer then Break;
						RefAccessor := RebuildingAccessor;
						case Decoder.Decode(RefAccessor) of
						#13, #10, UnicodeLineFeed, UnicodeFormFeed:
							{Ԃ牺};
						else
							Break
						end;
					end
				end;

				if not IsAlphabet(ExpellingLetter) or not IsAlphabet(Letter) then begin
					ExpellingX := PreWidth;
					ExpellingLetter := Letter;
					WrapX := PreWidth;
					WrapPos := PreAccessor.Index;
				end;

				if IndentingFlag and HasWideChar(Indenting, Letter) then begin
					IndentingFlag := Width <= MaxIndentWidth;
					if IndentingFlag then IndentWidth := Width;
				end else begin
					IndentingFlag := False;
				end
			end;

			if LineWidth > 0 then begin
				if RebuildingAccessor.EndOfBuffer then begin
					{s}[N\}
					if Width + StdWidth > LineWidth then begin
						MakeLine(RebuildingAccessor.Index);
					end;
				end else begin
					{܂Ԃ}
					if WordWrap and (Width - ExpellingX <= StdWidth * 4) and
						(HasWideChar(Expelling, ExpellingLetter) or IsAlphabet(ExpellingLetter)) then
					begin
						MakeLine(WrapPos);
						WrapWidth := Width - WrapX;
					end else begin
						MakeLine(RebuildingAccessor.Index);
						WrapWidth := 0;
					end
				end
			end;
	NewLine:
			if Width > FMaxWidth then FMaxWidth := Width;
		end;
		Result := RebuildingAccessor.Index;
	end;

var
	RepaintingFirst, RepaintingLast: Natural;
	I: Integer;
begin
	RepaintingFirst := First;
	RepaintingLast := RebuildLines;

	{}
	Analyze(RepaintingFirst, RepaintingLast);

	{ʒm}
	for I := High(FViews) downto Low(FViews) do begin
		FViews[I].AfterChange(First, OldLast, NewLast, RepaintingFirst, RepaintingLast, FLineModified);
	end;

	FLineModified := False;
end;

procedure TLayout.Analyze(var First, Last: Natural);
var
	I: Integer;
begin
	for I := First to Last - 1 do
		Sub[I] := Sub[I] and not SubStateMask or SubStateBase;
end;

procedure TLayout.BeforeChange(First, Last: Natural);
var
	I: Integer;
begin
	for I := First to Last - 1 do
		if Sub[I] and SubLineEnd <> 0 then
		begin
			Dec(FTotalLineCount);
			FLineModified := True;
		end;

	for I := High(FViews) downto Low(FViews) do
		FViews[I].BeforeChange(First, Last);
end;

constructor TLayout.Create(ADocument: TDocument; ADecoder: TDecoder;
	ALineWidth, ATabWidth, AFixedLetterWidth, ALineHeight: Integer;
	AHanging, AExpelling, AIndenting: WideString; AWordWrap: Boolean);
begin
	inherited Create;

	FDocument := ADocument;
	FDecoder := ADecoder;
	FLineWidth := ALineWidth;
	FTabWidth := ATabWidth;
	FFixedLetterWidth := AFixedLetterWidth;
	FLineHeight := ALineHeight;
	FHanging := AHanging;
	FExpelling := AExpelling;
	FIndenting := AIndenting;
	FWordWrap := AWordWrap;

	FMaxIndentWidth := ALineWidth div 2;

	ADocument.AddLayout(Self);
	FSub := ADocument.AllocateSubBuffer;

	FTotalLineCount := 1;

	BeforeChange(0, 0);
	AfterChange(0, 0, ADocument.Size);
end;

destructor TLayout.Destroy;
begin
	FDocument.RemoveLayout(Self);
	FDocument.ReleaseSubBuffer(FSub);
	inherited;
end;

function TLayout.GetLines(IncludingEnd: Boolean): TLines;
var
	Line: TLine;
	Count: Integer;
begin
	SetLength(Result, TotalLineCount + 1);
	Count := 0;
	Line := 0;
	while Line < Document.Size do
	begin
		Result[Count] := Line;
		Inc(Count);
		Line := LogicalSuccLine(Line);
		if Line < 0 then Break;
	end;
	if IncludingEnd then
	begin
		Result[Count] := Document.Size;
		Inc(Count);
	end;
	SetLength(Result, Count);
end;

function TLayout.GetLogicalLineEnd(Position: Natural): TLine;
begin
	Result := LogicalSuccLine(Position);
	if Result < 0 then
		Result := Document.Size
	else
		Result := PredLetter(Result)
end;

function TLayout.GetLogicalLineFromLineNumber(LineNumber: Integer): TLine;
var
	I: Integer;
begin
	Result := 0;
	for I := 2 to LineNumber do
		Result := LogicalSuccLine(Result);
end;

function TLayout.GetLogicalLineTop(Position: Natural): TLine;
begin
	if IsLogicalLineTop(Position) then
		Result := Position
	else
		Result := LogicalPredLine(Position)
end;

function TLayout.GetSub(AIndex: Natural): Byte;
begin
	Result := Document.Sub[SubIndex, AIndex]
end;

function TLayout.GetTabWidth(X: Integer): Integer;
var
	S: Integer;
begin
	S := TabWidth;
	Result := S - X mod S;
	if Result < StandardLetterWidth then Inc(Result, S)
end;

function TLayout.IsLogicalLineTop(Position: Natural): Boolean;
var
	Accessor: TBufferAccessor;
begin
	if Position = 0 then
		Result := True
	else if Sub[Position - 1] and SubLineEnd = 0 then
		Result := False
	else
	begin
		Accessor.Buffer := Document;
		Accessor.Index := PredLetter(Position);
		Result := Decoder.LineFeed(Accessor);
	end;
end;

function TLayout.LetterCount(Position: Natural): Natural;
var
	P: Natural;
begin
	Result := 0;
	P := Position;
	repeat
		Inc(Result);
		Inc(P);
		if P = Document.Size then Break;
		if Sub[P] and SubLetter <> 0 then Break;
	until False;
end;

function TLayout.LetterTop(Position: Natural): Natural;
begin
	Result := Position;
	if Result < Document.Size then
		repeat
			if Result = 0 then Break;
			if Sub[Result] and SubLetter <> 0 then Break;
			Dec(Result);
		until False;
end;

function TLayout.LogicalPredLine(Line: TLine): TLine;
var
	Accessor: TBufferAccessor;
begin
	if Line = 0 then
		Result := -1
	else
	begin
		Result := Line - 1;
		while Result > 0 do
		begin
			repeat
				Dec(Result);
				if Result = 0 then Exit;
				if Sub[Result] and SubLineEnd <> 0 then Break;
			until False;
			Accessor.Buffer := Document;
			Accessor.Index := LetterTop(Result);
			if Decoder.LineFeed(Accessor) then
			begin
				Inc(Result);
				Break;
			end;
			Dec(Result);
		end;
	end
end;

function TLayout.LogicalSuccLine(Line: TLine): TLine;
var
	Accessor: TBufferAccessor;
begin
	Accessor.Buffer := Document;
	Accessor.Index := Line;
	while not Accessor.EndOfBuffer do
	begin
		if Decoder.LineFeed(Accessor) then
		begin
			Result := SuccLetter(Accessor.Index);
			Exit;
		end;
		Accessor.Index := SuccLetter(Accessor.Index);
	end;
	Result := -1
end;

function TLayout.PredLetter(Position: Natural): Natural;
begin
	Result := Position;
	repeat
		if Result = 0 then Break;
		Dec(Result);
		if Sub[Result] and SubLetter <> 0 then Break;
	until False;
end;

function TLayout.PredLine(Line: TLine): TLine;
begin
	if Line = 0 then
		Result := -1
	else
	begin
		Result := Line;
		repeat
			Dec(Result);
		until (Result = 0) or (Sub[Result - 1] and SubLineEnd <> 0);
	end
end;

procedure TLayout.RemoveView(const View: TView);
var
	I, J: Integer;
begin
	for I := High(FViews) downto Low(FViews) do
		if FViews[I] = View then
		begin
			for J := I to High(FViews) - 1 do FViews[J] := FViews[J + 1];
			SetLength(FViews, Length(FViews) - 1);
			Break;
		end;
end;

procedure TLayout.SetSub(AIndex: Natural; const Value: Byte);
begin
	Document.Sub[SubIndex, AIndex] := Value
end;

function TLayout.SuccLetter(Position: Natural): Natural;
begin
	Result := Position;
	if Result < Document.Size then
		repeat
			Inc(Result);
			if Result = Document.Size then Break;
			if Sub[Result] and SubLetter <> 0 then Break;
		until False;
end;

function TLayout.SuccLine(Line: TLine): TLine;
begin
	if Line >= Document.Size then
		Result := -1
	else
	begin
		Result := Line;
		repeat
			if Sub[Result] and SubLineEnd <> 0 then
			begin
				Inc(Result);
				Exit;
			end;
			Inc(Result);
		until Result >= Document.Size;
		Result := -1
	end
end;

function TLayout.ViewCount: Integer;
begin
	Result := Length(FViews)
end;

{ TView }

procedure TView.AdjustRepaint(Base: TScrollBase; From: Natural;
	First, Last: Natural; out Y1, Y2: Integer);
var
	LL, Next, Line: TLine;
	LineHeight, Y: Integer;
	ViewRect: TRect;
	DocSize: Natural;
begin
	ViewRect := FViewUI.ViewRect;
	LineHeight := Layout.LineHeight;
	DocSize := Layout.Document.Size;

	if FViewUI.ScrollBase = sbTop then
	begin
		Y1 := Low(Integer);
		Y2 := Low(Integer);
		Y := ViewRect.Top;
		Line := From;
		repeat
			Next := SuccLine(Line);
			LL := Next;
			if LL < 0 then LL := DocSize;
			if (Y1 = Low(Integer)) and (LL >= First) then
			begin
				Y1 := Y;
				if (Line > 0) and not Layout.IsLogicalLineTop(Line) then
					Dec(Y1, LineHeight);
			end;
			if (Y2 = Low(Integer)) and (LL > Last) then
			begin
				Y2 := Y + LineHeight;
				if (Line < DocSize) and (LL = DocSize) then
					Inc(Y2, LineHeight);
				Break;
			end;
			Inc(Y, LineHeight);
			if Y >= ViewRect.Bottom then Break;
			Line := Next;
		until Line < 0;
		if (Y1 <> Low(Integer)) and (Y2 = Low(Integer)) then Y2 := ViewRect.Bottom;
	end
	else
	begin
		Y1 := Low(Integer);
		Y2 := Low(Integer);
		Y := ViewRect.Bottom;
		Line := From;
		repeat
			LL := Line;
			if (Y2 = Low(Integer)) and (LL <= Last) then
			begin
				Y2 := Y;
			end;
			if (Y1 = Low(Integer)) and (LL < First) then
			begin
				Y1 := Y - LineHeight;
				Break;
			end;
			Dec(Y, LineHeight);
			if Y < ViewRect.Top then Break;
			Line := PredLine(Line);
		until Line < 0;
		if (Y2 <> Low(Integer)) and (Y1 = Low(Integer)) then Y1 := ViewRect.Top;
	end
end;

procedure TView.AdjustScroll(var Left: Integer; var Base: TScrollBase; var From: Natural; out ScrollX, ScrollY: Integer);
var
	ViewRect: TRect;
	Y, LineHeight: Integer;
	Line, Next: TLine;
	HorzPos, HorzMin, HorzMax, HorzPage: Integer;
begin
	ScrollX := 0;
	if Left > 0 then
	begin
		HorzScrollBar(Left, HorzPos, HorzMin, HorzMax, HorzPage);
		if HorzMin < HorzMax then
		begin
			ScrollX := HorzPos + HorzPage - (HorzMax - HorzMin + 1);
			if ScrollX > 0 then
				Dec(Left, ScrollX)
			else
				ScrollX := 0;
		end;
	end;

	ViewRect := FViewUI.ViewRect;
	LineHeight := Layout.LineHeight;

	ScrollY := 0;
	if Base = sbTop then
	begin
		Y := ViewRect.Top + LineHeight;
		Line := From;
		while Line < Layout.Document.Size do
		begin
			Line := SuccLine(Line);
			if Line < 0 then Break;
			Inc(Y, LineHeight);
		end;
		if (From > 0) and (Y < ViewRect.Bottom) then
		begin
			Base := sbBottom;
			From := ToLine(Layout.Document.Size, +1);
    	ScrollY := ViewRect.Bottom - Y;
		end;
	end;
	if Base = sbBottom then
	begin
		if Layout.Document.Size < From then
		begin
			Y := 0;
			Line := From;
			while Layout.Document.Size < Line do
			begin
				Line := PredLine(Line);
				Inc(Y, LineHeight);
			end;
			From := Line;
			Inc(ScrollY, Y);
			Exit;
		end;

		Y := ViewRect.Bottom - LineHeight;
		Line := From;

		repeat
			Next := PredLine(Line);
			if Next < 0 then Break;
			Line := Next;
			Dec(Y, LineHeight);
			if Y <= ViewRect.Top then Exit;
		until False;

		Base := sbTop;
		From := 0;

		Inc(ScrollY, ViewRect.Top - Y);
	end
end;

function TView.BasedScrollUp(var Base: TScrollBase; var From: Natural): Integer;
var
	ViewRect: TRect;
	Line, Next: TLine;
begin
	if Base = sbTop then begin
		Result := ScrollUp(Base, From)
	end else begin
		ViewRect := FViewUI.ViewRect;
		Line := From;
		Result := Layout.LineHeight - (ViewRect.Bottom - ViewRect.Top);
		while Result <= 0 do begin
			Next := PredLine(Line);
			if Next < 0 then begin
				Base := sbTop;
				From := 0;
				Exit;
			end;
			Line := Next;
			Inc(Result, Layout.LineHeight);
		end;
		if Line > 0 then
		begin
			From := PredLine(From);
			Result := Layout.LineHeight;
    end
		else
		begin
			Base := sbTop;
			From := 0;
		end;
	end
end;

function TView.BasedScrollDown(var Base: TScrollBase; var From: Natural): Integer;
var
	ViewRect: TRect;
	Line, Next: TLine;
begin
	if Base = sbTop then begin
		ViewRect := FViewUI.ViewRect;
		Line := From;
		Result := (ViewRect.Bottom - ViewRect.Top) - Layout.LineHeight;
		while Result >= 0 do begin
			Next := SuccLine(Line);
			if Next < 0 then begin
				Result := 0;
				Exit;
			end;
			Line := Next;
			Dec(Result, Layout.LineHeight);
		end;
		if SuccLine(Line) >= 0 then begin
			From := SuccLine(From);
			Result := - Layout.LineHeight;
		end else begin
			Base := sbBottom;
			From := Line;
		end;
	end else begin
		Result := ScrollDown(Base, From);
	end;
end;

procedure TView.BeforeChange(First, Last: Natural);
begin
end;

constructor TView.Create(ALayout: TLayout; const AViewUI: IViewUI);
begin
	inherited Create;
	FLayout := ALayout;
	FViewUI := AViewUI;
	ALayout.AddView(Self);
end;

function TView.DeleteSingleLetter(Position: Natural; HexMode: Boolean; Dir: Integer): Natural;
var
	B: Integer;
begin
	Assert((Dir = 1) or (Dir = -1));
	Result := Position;
	if (Dir > 0) and (Result < Layout.Document.Size) or
		(Dir < 0) and (Result > 0) then
	begin
		B := Dir;
		if not HexMode then
		begin
			if Dir > 0 then
				B := Layout.LetterCount(Result)
			else
				B := - Layout.LetterCount(Layout.LetterTop(Result - 1))
		end;
		Result := InsertText(Result, B, '', HexMode)
	end;
end;

destructor TView.Destroy;
begin
	FLayout.RemoveView(Self);
	inherited;
end;

function TView.DownCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural;
var
	FromLine, NextLine: TLine;
begin
	if Layout.Document.Size = 0 then
		Result := 0
	else
	begin
		FromLine := ToLine(Position, +1);
		NextLine := SuccLine(FromLine);
		if NextLine < 0 then
			Result := Position
		else
		begin
			if X < 0 then X := ToX(FromLine, Position);
			Result := ToPosition(NextLine, X);
		end;
	end
end;

function TView.FixCaretPosition(Position: Natural): Natural;
begin
	Result := Layout.LetterTop(Position)
end;

function TView.InsertData(Position: Natural; DeleteLength: Integer; const Data: string; HexMode: Boolean): Natural;
begin
	Layout.Document.Replace(Position, DeleteLength, - Length(Data), Data[1]);

	Result := Position + Length(Data);
	if DeleteLength < 0 then Inc(Result, DeleteLength);
	if not HexMode then
		Result := FixCaretPosition(Result);
end;

function TView.InsertText(Position: Natural; DeleteLength: Integer; const Data: WideString; HexMode: Boolean): Natural;
begin
	Result := InsertData(Position, DeleteLength, Layout.Decoder.EncodeString(Data), HexMode)
end;

function TView.LineEnd(Position: Natural): Natural;
var
	N: Integer;
begin
	N := SuccLine(Position);
	if N < 0 then
		Result := Layout.Document.Size
	else
		Result := N
end;

function TView.ReadData(First, Last: Natural): string;
var
	Length: Natural;
begin
	Assert(First <= Last);
	Length := Last - First;
	SetLength(Result, Length);
	Layout.Document.Read(First, Length, Pointer(Result)^);
end;

function TView.ReadSingleLetter(Position: Natural): TLetterBinaryCode;
var
	Accessor1, Accessor2: TBufferAccessor;
	I: Integer;
begin
	if Position = Layout.Document.Size then
		Result.Length := 0
	else
	begin
		Accessor1.Buffer := Layout.Document;
		Accessor1.Index := Position;
		Accessor2 := Accessor1;
		Layout.Decoder.Decode(Accessor2);
		Result.Length := Accessor2.Index - Accessor1.Index;
		for I := 0 to Result.Length - 1 do
			Result.Codes[I] := Accessor1.Take;
	end;
end;

function TView.ReadText(First, Last: Natural): WideString;
var
	Accessor: TBufferAccessor;
begin
	Assert(First <= Last);
	Accessor.Buffer := Layout.Document;
	Accessor.Index := First;
	Result := '';
	while Accessor.Index < Last do
		Result := Result + Layout.Decoder.Decode(Accessor)
end;

function TView.ScrollDown(var Base: TScrollBase; var From: Natural): Integer;
var
	ViewRect: TRect;
	Line, Next: TLine;
begin
	if Base = sbTop then
	begin
		ViewRect := FViewUI.ViewRect;
		Line := From;
		Result := (ViewRect.Bottom - ViewRect.Top) - Layout.LineHeight;
		while Result >= 0 do
		begin
			Next := SuccLine(Line);
			if Next < 0 then
			begin
				Result := 0;
				Exit;
			end;
			Line := Next;
			Dec(Result, Layout.LineHeight);
		end;
		Base := sbBottom;
		From := Line;
	end
	else
	begin
		Next := SuccLine(From);
		if Next < 0 then
			Result := 0
		else
		begin
			From := Next;
			Result := - Layout.LineHeight;
		end
	end;
end;

function TView.ScrollUp(var Base: TScrollBase; var From: Natural): Integer;
var
	ViewRect: TRect;
	Line, Next: TLine;
begin
	if Base = sbTop then
	begin
		Next := PredLine(From);
		if Next < 0 then
			Result := 0
		else
		begin
			From := Next;
			Result := Layout.LineHeight;
		end
	end
	else 
	begin
		ViewRect := FViewUI.ViewRect;
		Line := From;
		Result := Layout.LineHeight - (ViewRect.Bottom - ViewRect.Top);
		while Result <= 0 do
		begin
			Next := PredLine(Line);
			if Next < 0 then Break;
			Line := Next;
			Inc(Result, Layout.LineHeight);
		end;
		Base := sbTop;
		From := Line;
	end
end;

function TView.UpCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural;
var
	FromLine, NextLine: TLine;
begin
	if Layout.Document.Size = 0 then
		Result := 0
	else
	begin
		FromLine := ToLine(Position, +1);
		NextLine := PredLine(FromLine);
		if NextLine < 0 then
			Result := Position
		else
		begin
			if X < 0 then X := ToX(FromLine, Position);
			Result := ToPosition(NextLine, X);
		end;
	end
end;

procedure TView.VertScroll(var Base: TScrollBase; var From: Natural;
	Position, Min, Max, PageSize: Integer);
var
	L: Integer;
	Next: TLine;
begin
	L := Position;
	if L + PageSize >= Max then
	begin
		Base := sbBottom;
		From := ToLine(Layout.Document.Size, +1);
	end
	else
	begin
		Base := sbTop;
		From := 0;
		while L > 0 do
		begin
			Next := SuccLine(From);
			if Next < 0 then Break;
			From := Next;
			Dec(L);
		end;
	end;
end;

procedure TView.VertScrollBarInternal(LineCount, LinePosition: Integer;
	out Position, Min, Max, PageSize: Integer);
var
	ViewRect: TRect;
	Base: TScrollBase;
begin
	ViewRect := FViewUI.ViewRect;
	Base := FViewUI.ScrollBase;

	PageSize := (ViewRect.Bottom - ViewRect.Top) div Layout.LineHeight;

	if PageSize >= LineCount then
	begin
    Position := 0;
		Min := 0;
		Max := 0;
    PageSize := 0;
	end
	else
	begin
		Min := 0;
		Max := LineCount - 1;
		if Base = sbTop then
		begin
			Position := LinePosition;
			if Position + PageSize > Max then PageSize := Max - Position;
		end
		else
		begin
			Position := 1 + LinePosition - PageSize;
			if Position < 0 then Position := 0;
		end
	end
end;

function TView.ViewType: TViewClass;
begin
	Result := TViewClass(ClassType)
end;

{ TTextView }

procedure TTextView.AdjustScrollOnChange(First, OldLast, NewLast: Natural;
	var Base: TScrollBase; var From: Natural);
var
	A, B: TLine;
begin
	if (From > First) and (From <= OldLast) then
	begin
		if OldLast <= NewLast then
			From := ToLine(From, +1)
		else if Base = sbTop then
			From := ToLine(First, +1)
		else
			From := ToLine(NewLast, +1)
	end
	else if From > OldLast then
	begin
		Inc(From, NewLast - OldLast);
		{Ȃ}
		A := ToLine(From, +1);
		B := SuccLine(A); if B < 0 then B:= A;
		if Abs(A - From) < Abs(B - From) then From := A else From := B;
	end;
end;

procedure TTextView.AfterChange(First, OldLast, NewLast,
	RepaintingFirst, RepaintingLast: Natural; LineModified: Boolean);
begin
	if LineModified then
	begin
		if FViewUI.ScrollBase = sbTop then
			RepaintingLast := Layout.Document.Size
		else
			RepaintingFirst := 0;
	end;

	FViewUI.Changed(First, OldLast, NewLast, RepaintingFirst, RepaintingLast)
end;

function TTextView.AutoIndent(Position: Natural): WideString;
var
	Line: TLine;
	Accessor: TBufferAccessor;
	Letter: WideChar;
begin
	Result := '';

	Line := Layout.GetLogicalLineTop(Position);
	Accessor.Buffer := Layout.Document;
	Accessor.Index := Line;

	while not Accessor.EndOfBuffer do
	begin
		if Accessor.Index >= Position then Break;
		Letter := Layout.Decoder.Decode(Accessor);
		if not HasWideChar(Layout.Indenting, Letter) then Break;
		Result := Result + Letter;
	end;
end;

procedure TTextView.BeforeChange(First, Last: Natural);
begin
	while FLineNumberFrom > First do
	begin
		if Layout.IsLogicalLineTop(FLineNumberFrom) then Dec(FLineNumber);
		FLineNumberFrom := PredLine(FLineNumberFrom);
		Dec(FScreenNumber);
	end;
	Assert(FLineNumberFrom >= 0);
end;

constructor TTextView.Create(ALayout: TLayout; const AViewUI: IViewUI);
begin
	FLineNumberFrom := 0;
	FLineNumber := 1;
	FScreenNumber := 0;
	inherited;
end;

procedure TTextView.GetLocation(Position: Natural; out ScreenNumber, ScreenColumn, LineNumber, Column: Integer);
var
	TargetLine: TLine;
begin
	TargetLine := ToLine(Position, +1);
	if TargetLine > FLineNumberFrom then
	begin
		repeat
			FLineNumberFrom := SuccLine(FLineNumberFrom);
			Inc(FScreenNumber);
			if Layout.IsLogicalLineTop(FLineNumberFrom) then Inc(FLineNumber);
			Assert(FLineNumberFrom >= 0);
		until FLineNumberFrom = TargetLine;
	end
	else if TargetLine < FLineNumberFrom then
	begin
		repeat
			if Layout.IsLogicalLineTop(FLineNumberFrom) then Dec(FLineNumber);
			FLineNumberFrom := PredLine(FLineNumberFrom);
			Dec(FScreenNumber);
			Assert(FLineNumberFrom >= 0);
		until FLineNumberFrom = TargetLine;
	end;

	ScreenNumber := FScreenNumber;
	ScreenColumn := Position - TargetLine + 1;
	LineNumber := FLineNumber;
	Column := Position - Layout.GetLogicalLineTop(TargetLine) + 1;
end;

procedure TTextView.HorzScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer);
var
	ViewRect: TRect;
begin
	ViewRect := FViewUI.ViewRect;
	Position := ScrollFrom;
	Min := 0;
	Max := Layout.FMaxWidth + Layout.StandardLetterWidth * 16;
	PageSize := ViewRect.Right - ViewRect.Left;
	if PageSize > Max then
		PageSize := Max + 1;
end;

function TTextView.Indent(First, Last: Natural): Natural;
var
	Line: TLine;
	S: string;
begin
	S := Layout.Decoder.EncodeString(#9);
	Result := Last;
	Line := ToLine(Last, -1);
	while (Line >= 0) and (Line >= First) do
	begin
		if Line = Layout.GetLogicalLineTop(Line) then
		begin
			InsertData(Line, 0, S, False);
			Inc(Result, Length(S));
		end;
		Line := PredLine(Line);
	end;
end;

function TTextView.IndentWidth(Line: TLine): Integer;
var
	P: TLine;
	Accessor: TBufferAccessor;
	Letter: WideChar;
	Width: Integer;
begin
	P := Layout.GetLogicalLineTop(Line);
	Result := 0;
	if P < Line then
	begin
		Accessor.Buffer := Layout.Document;
		Accessor.Index := P;
		while not Accessor.EndOfBuffer do
		begin
			Letter := Layout.Decoder.Decode(Accessor);
			if HasWideChar(Layout.Indenting, Letter) then
			begin
				if Letter = #9 then
					Width := TabWidth(Result)
				else
					Width := Layout.LetterWidth(Letter);

				if Result + Width > Layout.MaxIndentWidth then Break;
				Inc(Result, Width);
			end
			else
				Break
		end
	end;
end;

function TTextView.LeftCaretPosition(HexMode: Boolean; Position: Natural): Natural;
begin
	Assert(not HexMode);
	Result := Layout.PredLetter(Position)
end;

function TTextView.LineEndCaretPosition(Position: Natural): Natural;
begin
	Result := Layout.GetLogicalLineEnd(Position)
end;

function TTextView.LineTopCaretPosition(Position: Natural; Smart: Boolean): Natural;
var
	IndentedPosition: Natural;
	Accessor: TBufferAccessor;
begin
	Result := Layout.GetLogicalLineTop(Position);
	if Smart then
	begin
		IndentedPosition := Result;
		Accessor.Buffer := Layout.Document;
		Accessor.Index := Result;
		while not Accessor.EndOfBuffer do
		begin
			if not HasWideChar(Layout.Indenting, Layout.Decoder.Decode(Accessor)) then
				Break;
			IndentedPosition := Accessor.Index;
		end;

		if (Position > IndentedPosition) or (Position = Result) then
			Result := IndentedPosition;
	end;
end;

function TTextView.PredLine(Position: Natural): TLine;
begin
	Result := Layout.PredLine(Position)
end;

procedure TTextView.ReadyPaintLine(Base: TScrollBase; From: Natural;
	out Y: Integer; out NewLine: Boolean;
	out LineNumber, Indent: Integer; out Accessor: TBufferAccessor);
var
	Line, Next: TLine;
	ViewRect: TRect;
	LineHeight, DummyScreenNumber, DummyScreenColumn, DummyColumn: Integer;
begin
	ViewRect := FViewUI.ViewRect;
	if Base = sbTop then
	begin
		Line := From;
		Y := ViewRect.Top;
	end
	else
	begin
		LineHeight := Layout.LineHeight;
		Line := From;
		Y := ViewRect.Bottom;
		Dec(Y, LineHeight);
		while Y > ViewRect.Top do
		begin
			Next := PredLine(Line);
			if Next < 0 then Break;
			Line := Next;
			Dec(Y, LineHeight);
		end
	end;

	NewLine := Layout.IsLogicalLineTop(Line);
	GetLocation(Line, DummyScreenNumber, DummyScreenColumn, LineNumber, DummyColumn);
	Indent := IndentWidth(Line);
	Accessor.Buffer := Layout.Document;
	Accessor.Index := Line;
end;

function TTextView.RightCaretPosition(HexMode: Boolean; Position: Natural): Natural;
begin
	Assert(not HexMode);
	Result := Layout.SuccLetter(Position)
end;

function TTextView.SuccLine(Position: Natural): TLine;
begin
	Result := Layout.SuccLine(Position)
end;

function TTextView.TabWidth(X: Integer): Integer;
begin
	Result := Layout.GetTabWidth(X)
end;

function TTextView.ToLine(Position: Natural; Dir: Integer): TLine;
begin
	Result := Position;
	if Result > 0 then
	begin
		if Layout.Sub[Result - 1] and SubLineEnd <> 0 then
		begin
			if Dir < 0 then Result := PredLine(Result);
			Exit;
		end;
		Dec(Result);
		while (Result > 0) and (Layout.Sub[Result - 1] and SubLineEnd = 0) do
		begin
			Dec(Result);
		end
	end
end;

function TTextView.ToPosition(Line: TLine; X: Integer): Natural;
var
	CurX, W: Integer;
	Accessor: TBufferAccessor;
	Letter: WideChar;
begin
	CurX := IndentWidth(Line);
	Accessor.Buffer := Layout.Document;
	Accessor.Index := Line;
	while not Accessor.EndOfBuffer do
	begin
		Result := Accessor.Index;
		Letter := Layout.Decoder.Decode(Accessor);
		if Layout.Sub[Accessor.Index - 1] and SubLineEnd <> 0 then Exit;
		if Letter = #9 then
			W := TabWidth(CurX)
		else
			W := Layout.LetterWidth(Letter);
		Inc(CurX, W);
		if (CurX > X + (W div 2)) then Exit;
	end;
	Result := Layout.Document.Size;
end;

function TTextView.ToX(Line: TLine; Position: Natural): Integer;
var
	Accessor: TBufferAccessor;
	Letter: WideChar;
begin
	Result := IndentWidth(Line);
	Accessor.Buffer := Layout.Document;
	Accessor.Index := Line;
	while Accessor.Index < Position do
	begin
		Letter := Layout.Decoder.Decode(Accessor);
		if Letter = #9 then
			Inc(Result, TabWidth(Result))
		else
			Inc(Result, Layout.LetterWidth(Letter));
	end;
end;

function TTextView.Unindent(First, Last: Natural): Natural;
var
	Line: TLine;
	S: Integer;
	Accessor: TBufferAccessor;
begin
	Result := Last;
	Line := ToLine(Last, -1);
	while (Line >= 0) and (Line >= First) do
	begin
		if (Line < Layout.Document.Size) and (Line = Layout.GetLogicalLineTop(Line)) then
		begin
			Accessor.Buffer := Layout.Document;
			Accessor.Index := Line;
			if Layout.Decoder.Decode(Accessor) = #9 then
			begin
				S := Accessor.Index - Line;
				InsertData(Line, S, '', False);
				Dec(Result, S);
			end;
		end;
		Line := PredLine(Line)
	end;
end;

procedure TTextView.VertScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer);
var
	LineCount, SN, SC, LN, CL: Integer;
begin
	GetLocation(ScrollFrom, SN, SC, LN, CL);
	LineCount := Layout.TotalLineCount;
	VertScrollBarInternal(LineCount, SN, Position, Min, Max, PageSize);
end;

{ TBinaryView }

procedure TBinaryView.AfterChange(First, OldLast, NewLast,
	RepaintingFirst, RepaintingLast: Natural; LineModified: Boolean);
begin
	RepaintingLast := Layout.Document.Size;
	FViewUI.Changed(First, OldLast, NewLast, RepaintingFirst, RepaintingLast)
end;

function TBinaryView.DownCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural;
begin
	if HexMode then
	begin
		Result := Position;
		Inc(Result, 16);
		if Result > Layout.Document.Size then
			Result := Layout.Document.Size;
	end
	else
		Result := inherited DownCaretPosition(HexMode, Position, X);
end;

procedure TBinaryView.HorzScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer);
begin
  Position := 0;
	Min := 0;
	Max := 0;
	PageSize := 0;
end;

function TBinaryView.LeftCaretPosition(HexMode: Boolean; Position: Natural): Natural;
begin
	if HexMode then
	begin
		Result := Position;
		if Result > 0 then
			Dec(Result);
	end
	else
		Result := Layout.PredLetter(Position)
end;

function TBinaryView.LineEndCaretPosition(Position: Natural): Natural;
begin
	Result := ToLine(Position, +1);
	if Result + 15 >= Layout.Document.Size then
		Result := Layout.Document.Size
	else
	begin
		Inc(Result, 15);
		Result := Layout.LetterTop(Result);
	end
end;

function TBinaryView.LineTopCaretPosition(Position: Natural; Smart: Boolean): Natural;
var
	P: Natural;
begin
	Result := ToLine(Position, +1);
	P := Layout.LetterTop(Result);
	if P < Result then
		Result := Layout.SuccLetter(P);
end;

function TBinaryView.PredLine(Position: Natural): TLine;
begin
	Assert(Position mod 16 = 0);
	if Position = 0 then
		Result := -1
	else if Position < 16 then
		Result := 0
	else
		Result := Position - 16;
end;

procedure TBinaryView.ReadyPaintLine(Base: TScrollBase; From: Natural;
	out Y: Integer; out NewLine: Boolean;
	out LineNumber, Indent: Integer; out Accessor: TBufferAccessor);
var
	Line: TLine;
	ViewRect: TRect;
	LineHeight, VisibleLines: Integer;
begin
	ViewRect := FViewUI.ViewRect;
	if Base = sbTop then
	begin
		Y := ViewRect.Top;
		Line := From;
	end
	else
	begin
		LineHeight := Layout.LineHeight;
		VisibleLines := (ViewRect.Bottom - ViewRect.Top + LineHeight - 1) div LineHeight;
		Y := ViewRect.Bottom - VisibleLines * LineHeight;
		Line := From - (VisibleLines - 1) * 16;
		if Line < 0 then
		begin
			Y := ViewRect.Top;
			Line := 0;
		end
  end;		
	NewLine := False;
	Indent := 0;
	LineNumber := 0;
	Accessor.Buffer := Layout.Document;
	Accessor.Index := Line;
end;

function TBinaryView.RightCaretPosition(HexMode: Boolean; Position: Natural): Natural;
begin
	if HexMode then
	begin
		Result := Position;
		if Result < Layout.Document.Size then
			Inc(Result);
	end
	else
		Result := Layout.SuccLetter(Position)
end;

function TBinaryView.SuccLine(Position: Natural): TLine;
begin
	Assert(Position mod 16 = 0);
	if Position = Layout.Document.Size then
		Result := -1
	else
	begin
		Result := Position + 16;
		if Result >= Layout.Document.Size then
		begin
			if Layout.Document.Size mod 16 = 0 then
				Result := Layout.Document.Size
			else
				Result := -1
		end;
	end
end;

function TBinaryView.ToLine(Position: Natural; Dir: Integer): TLine;
begin
	Result := (Position div 16) * 16;
	if (Result = Position) and (Dir < 0) then
		Result := PredLine(Result)
end;

function TBinaryView.ToPosition(Line: TLine; X: Integer): Natural;
var
	CurX, Bytes: Integer;
begin
	X := X div Layout.FixedLetterWidth;

	Result := Layout.LetterTop(Line);
	CurX := Result - Line;

	while Result < Layout.Document.Size do
	begin
		Bytes := Layout.LetterCount(Result);
		Inc(CurX, Bytes);
		if CurX > X + (Bytes div 2) then Break;
		Inc(Result, Bytes);
	end;

	if Result < Line then
		Result := Layout.SuccLetter(Result);
end;

function TBinaryView.ToX(Line: TLine; Position: Natural): Integer;
begin
	Result := Position - Line;
	if Result < 0 then Result := 0 else if Result > 15 then Result := 15;
	Result := Result * Layout.FixedLetterWidth;
end;

function TBinaryView.UpCaretPosition(HexMode: Boolean; Position: Natural; var X: Integer): Natural;
begin
	if HexMode then
	begin
		Result := Position;
		if Result >= 16 then
			Dec(Result, 16);
	end
	else
		Result := inherited UpCaretPosition(HexMode, Position, X)
end;

procedure TBinaryView.VertScrollBar(ScrollFrom: Integer; out Position, Min, Max, PageSize: Integer);
var
	LineCount: Integer;
begin
	LineCount := Layout.Document.Size div 16;
	if Layout.Document.Size mod 16 = 0 then
		Inc(LineCount);

	VertScrollBarInternal(LineCount, ScrollFrom div 16, Position, Min, Max, PageSize);
end;

end.

